# -*- coding: utf-8 -*-
# 2023/5/29
# create by: snower

from syncany.calculaters import typing_filter
from ...utils import NumberDecimalTypes, ensure_number, ensure_str

@typing_filter(int)
def mysql_eq(a, b):
    if a.__class__ is b.__class__:
        if a is None:
            return None
        return 1 if a == b else 0
    if a is None or b is None:
        return None
    if isinstance(a, NumberDecimalTypes):
        if isinstance(b, NumberDecimalTypes):
            return 1 if a == b else 0
        try:
            return 1 if ensure_number(a) == ensure_number(b) else 0
        except:
            return 1 if a == b else 0
    if isinstance(a, str):
        if isinstance(b, str):
            return 1 if a == b else 0
        return 1 if ensure_str(a) == ensure_str(b) else 0
    if isinstance(a, bool):
        if isinstance(b, bool):
            return 1 if a == b else 0
        try:
            return 1 if ensure_number(a) == ensure_number(b) else 0
        except:
            return 1 if a == b else 0
    return 1 if a == b else 0

@typing_filter(int)
def mysql_neq(a, b):
    if a.__class__ is b.__class__:
        if a is None:
            return None
        return 1 if a != b else 0
    if a is None or b is None:
        return None
    if isinstance(a, NumberDecimalTypes):
        if isinstance(b, NumberDecimalTypes):
            return 1 if a != b else 0
        try:
            return 1 if ensure_number(a) != ensure_number(b) else 0
        except:
            return 1 if a != b else 0
    if isinstance(a, str):
        if isinstance(b, str):
            return 1 if a != b else 0
        return 1 if ensure_str(a) != ensure_str(b) else 0
    if isinstance(a, bool):
        if isinstance(b, bool):
            return 1 if a != b else 0
        try:
            return 1 if ensure_number(a) != ensure_number(b) else 0
        except:
            return 1 if a != b else 0
    return 1 if a != b else 0

@typing_filter(int)
def mysql_gt(a, b):
    if a.__class__ is b.__class__:
        if a is None:
            return None
        try:
            return 1 if a > b else 0
        except:
            pass
    if a is None or b is None:
        return None
    if isinstance(a, NumberDecimalTypes):
        if isinstance(b, NumberDecimalTypes):
            return 1 if a > b else 0
        try:
            return 1 if ensure_number(a) > ensure_number(b) else 0
        except:
            return 1 if a > b else 0
    if isinstance(a, str):
        if isinstance(b, str):
            return 1 if a > b else 0
        return 1 if ensure_str(a) > ensure_str(b) else 0
    if isinstance(a, bool):
        if isinstance(b, bool):
            return 1 if a > b else 0
        try:
            return 1 if ensure_number(a) > ensure_number(b) else 0
        except:
            return 1 if a > b else 0
    return 1 if a > b else 0

@typing_filter(int)
def mysql_gte(a, b):
    if a.__class__ is b.__class__:
        if a is None:
            return None
        try:
            return 1 if a >= b else 0
        except:
            pass
    if a is None or b is None:
        return None
    if isinstance(a, NumberDecimalTypes):
        if isinstance(b, NumberDecimalTypes):
            return 1 if a >= b else 0
        try:
            return 1 if ensure_number(a) >= ensure_number(b) else 0
        except:
            return 1 if a >= b else 0
    if isinstance(a, str):
        if isinstance(b, str):
            return 1 if a >= b else 0
        return 1 if ensure_str(a) >= ensure_str(b) else 0
    if isinstance(a, bool):
        if isinstance(b, bool):
            return 1 if a >= b else 0
        try:
            return 1 if ensure_number(a) >= ensure_number(b) else 0
        except:
            return 1 if a >= b else 0
    return 1 if a >= b else 0

@typing_filter(int)
def mysql_lt(a, b):
    if a.__class__ is b.__class__:
        if a is None:
            return None
        try:
            return 1 if a < b else 0
        except:
            pass
    if a is None or b is None:
        return None
    if isinstance(a, NumberDecimalTypes):
        if isinstance(b, NumberDecimalTypes):
            return 1 if a < b else 0
        try:
            return 1 if ensure_number(a) < ensure_number(b) else 0
        except:
            return 1 if a < b else 0
    if isinstance(a, str):
        if isinstance(b, str):
            return 1 if a < b else 0
        return 1 if ensure_str(a) < ensure_str(b) else 0
    if isinstance(a, bool):
        if isinstance(b, bool):
            return 1 if a < b else 0
        try:
            return 1 if ensure_number(a) < ensure_number(b) else 0
        except:
            return 1 if a < b else 0
    return 1 if a < b else 0

@typing_filter(int)
def mysql_lte(a, b):
    if a.__class__ is b.__class__:
        if a is None:
            return None
        try:
            return 1 if a <= b else 0
        except:
            pass
    if a is None or b is None:
        return None
    if isinstance(a, NumberDecimalTypes):
        if isinstance(b, NumberDecimalTypes):
            return 1 if a <= b else 0
        try:
            return 1 if ensure_number(a) <= ensure_number(b) else 0
        except:
            return 1 if a <= b else 0
    if isinstance(a, str):
        if isinstance(b, str):
            return 1 if a <= b else 0
        return 1 if ensure_str(a) <= ensure_str(b) else 0
    if isinstance(a, bool):
        if isinstance(b, bool):
            return 1 if a <= b else 0
        try:
            return 1 if ensure_number(a) <= ensure_number(b) else 0
        except:
            return 1 if a <= b else 0
    return 1 if a <= b else 0

@typing_filter(int)
def mysql_in(a, b):
    if a is None:
        return None
    hs_none = False
    for v in b:
        if v is None:
            hs_none = True
            continue
        if mysql_eq(a, v):
            return 1
    return None if hs_none else 0

@typing_filter(int)
def mysql_not(a):
    if a is None:
        return None
    return 1 if not a else 0

@typing_filter(int)
def mysql_is(a, b):
    if a.__class__ is b.__class__:
        return 1 if a is b else 0
    if a is None or b is None:
        return 1 if a is b else 0
    return mysql_eq(a, b)

@typing_filter(int)
def mysql_is_not(a, b):
    if a.__class__ is b.__class__:
        return 0 if a is b else 1
    if a is None or b is None:
        return 0 if a is b else 1
    return mysql_neq(a, b)

@typing_filter(int)
def mysql_exists(a):
    if a is None:
        return 0
    if a.__class__ is dict:
        return 1
    if isinstance(a, dict):
        return 1
    return 1 if a else 0


funcs = {key[6:]: value for key, value in globals().items() if key.startswith("mysql_")}


# -*- coding: utf-8 -*-
# 2024/2/2
# create by: snower

from syncany.calculaters.calculater import Calculater


class RowIndexCalculater(Calculater):
    @classmethod
    def instance(cls, name):
        return RowIndexCalculater(name)

    def __init__(self, *args, **kwargs):
        super(RowIndexCalculater, self).__init__(*args, **kwargs)

        self.row_index = 0

    def calculate(self, *args):
        self.row_index += 1
        return self.row_index


class RowLastCalculater(Calculater):
    @classmethod
    def instance(cls, name):
        return RowLastCalculater(name)

    def __init__(self, *args, **kwargs):
        super(RowLastCalculater, self).__init__(*args, **kwargs)

        self.row_last = None

    def calculate(self, *args):
        row_last, self.row_last = self.row_last, (args[0] if len(args) == 1 else args)
        return row_last



# -*- coding: utf-8 -*-
# 23/02/07
# create by: snower

from .version import version, version_info

import threading
from syncany.loaders import Loader, register_loader
from syncany.outputers import Outputer, register_outputer
from syncany.valuers import Valuer, register_valuer
from syncany.filters import Filter, register_filter
from syncany.database import DataBase, register_database
from syncany.calculaters import Calculater, TypeFormatCalculater, TypingCalculater, MathematicalCalculater, \
    TransformCalculater, register_calculater
from syncany.taskers.config import Parser, ConfigReader, register_parser, register_reader
from syncany.taskers.tasker import current_tasker
from syncany.taskers.manager import TaskerManager
from syncany.database import DatabaseManager, find_database
from syncany.errors import DatabaseUnknownException
from .version import version, version_info
from .parser import SqlParser, SqlSegment, FileParser
from .config import GlobalConfig
from .executor import Executor
from .calculaters import GenerateCalculater, AggregateCalculater, StateAggregateCalculater, \
    WindowAggregateCalculater, WindowStateAggregateCalculater


class ExecuterContext(object):
    _execute_index = 0
    _thread_local = threading.local()

    @classmethod
    def current(cls):
        return cls._thread_local.current_executer_context

    def __init__(self, engine, executor):
        self.engine = engine
        self.executor = executor

    def get_variable(self, name, default=None):
        if name and name[0] != "@":
            name = "@" + name
        return self.executor.env_variables.get(name, default)

    def set_variable(self, name, value):
        if name and name[0] != "@":
            name = "@" + name
        self.executor.env_variables[name] = value

    def use(self, name, func_or_module):
        self.executor.session_config.set(("imports", name), func_or_module)

    def execute(self, sql):
        sql_parser = SqlParser(sql)
        sqls = sql_parser.split()
        self.__class__._execute_index += 1
        with self.executor as executor:
            executor.run("execute[%d]" % self._execute_index, sqls)
            executor.execute()

    def get_database(self, db=None):
        return self.engine.get_database(db)

    def get_memory_datas(self, name):
        return self.engine.get_memory_datas(name)

    def pop_memory_datas(self, name):
        return self.engine.pop_memory_datas(name)

    def push_memory_datas(self, name, datas):
        return self.engine.push_memory_datas(name, datas)

    def terminal(self):
        if not self.executor:
            return
        self.executor.terminate()

    def __enter__(self):
        self._thread_local.current_executer_context = self
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._thread_local.current_executer_context = None


class ScriptEngine(object):
    default_instance = None

    @classmethod
    def instance(cls, config=None):
        if cls.default_instance is None:
            cls.default_instance = ScriptEngine(config)
        return cls.default_instance

    @classmethod
    def current_context(cls):
        return ExecuterContext.current()

    def __init__(self, custom_config=None):
        self.custom_config = custom_config
        self.config = GlobalConfig()
        self.manager = None
        self.databases = {}
        self.executor = None

    def setup(self):
        if self.manager is not None:
            return
        init_execute_files = self.config.load(self.custom_config)
        self.config.config_logging()
        self.config.load_extensions()
        self.manager = TaskerManager(DatabaseManager())
        self.executor = Executor(self.manager, self.config.session())
        if init_execute_files:
            for init_execute_file in init_execute_files:
                file_parser = FileParser(init_execute_file)
                self.executor.run("init " + init_execute_file, file_parser.load())
                with self.executor as executor:
                    executor.execute()

    def get_variable(self, name, default=None):
        if self.executor is None:
            self.setup()
        if name and name[0] != "@":
            name = "@" + name
        return self.executor.env_variables.get(name, default)

    def set_variable(self, name, value):
        if self.executor is None:
            self.setup()
        if name and name[0] != "@":
            name = "@" + name
        self.executor.env_variables[name] = value

    def use(self, name, func_or_module):
        if self.executor is None:
            self.setup()
        self.executor.session_config.set(("imports", name), func_or_module)

    def context(self):
        if self.executor is None:
            self.setup()
        return ExecuterContext(self, Executor(self.manager, self.executor.session_config.session(), self.executor))

    def execute(self, sql):
        if self.executor is None:
            self.setup()
        return ExecuterContext(self, self.executor).execute(sql)

    def get_database(self, db=None):
        if self.manager is None:
            self.setup()
        db = db if db else "--"
        if db in self.databases:
            return self.databases[db]
        try:
            database_config = dict(**[database for database in self.executor.session_config.get()["databases"]
                                      if database["name"] == db][0])
        except Exception:
            raise DatabaseUnknownException("%s is unknown" % db)
        database_cls = find_database(database_config.pop("driver"))
        if not database_cls:
            raise DatabaseUnknownException(database_config["name"] + " is unknown")
        self.databases[db] = database_cls(self.manager.database_manager, database_config).build()
        return self.databases[db]

    def get_memory_datas(self, name):
        if self.manager is None:
            return []
        database = self.get_database("--")
        query = database.query("--." + name, ["id"])
        return query.commit()

    def pop_memory_datas(self, name):
        if self.manager is None:
            return []
        database = self.get_database("--")
        query = database.query("--." + name, ["id"])
        datas = query.commit()
        delete = database.delete("--." + name, ["id"])
        delete.commit()
        return datas

    def push_memory_datas(self, name, datas):
        database = self.get_database("--")
        insert = database.insert("--." + name, ["id"], datas=datas if isinstance(datas, list) else [datas])
        insert.commit()

    def terminal(self):
        executer_context = ExecuterContext.current()
        if not executer_context:
            if not self.executor:
                return
            self.executor.terminate()
            return
        executer_context.terminate()

    def close(self):
        if self.manager:
            self.manager.close()
            self.manager = None
        if self.databases:
            for name, database in self.databases.items():
                database.close()
            self.databases = {}
        self.executor = None
        self.config = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()


def instance(config=None):
    return ScriptEngine.instance(config)


def get_variable(name, default=None):
    return ScriptEngine.instance().get_variable(name, default)


def set_variable(name, value):
    ScriptEngine.instance().set_variable(name, value)


def use(name, func_or_module):
    ScriptEngine.instance().use(name, func_or_module)


def execute(sql):
    return ScriptEngine.instance().execute(sql)


def get_database(db=None):
    return ScriptEngine.instance().get_database(db)


def get_memory_datas(name):
    return ScriptEngine.instance().get_memory_datas(name)


def pop_memory_datas(name):
    return ScriptEngine.instance().pop_memory_datas(name)


def push_memory_datas(name, datas):
    return ScriptEngine.instance().push_memory_datas(name, datas)


def terminal():
    return ScriptEngine.instance().terminal()


def close():
    if ScriptEngine.default_instance is None:
        return
    ScriptEngine.default_instance.close()
    ScriptEngine.default_instance = None



# -*- coding: utf-8 -*-
# 2023/9/27
# create by: snower

from .example import ExampleTestCase


class LogicOperationExampleTestCase(ExampleTestCase):
    example_name = "logic_operation"

    def test_logic_operation(self):
        self.execute("logic_operation.sql")

        self.assert_data(2, [{'a': 1, 'b': 1, 'c': 0, 'd': 1, 'f': 1}], "data error")

        self.assert_data(4, [{'a': 1, 'b': 0}], "data error")

        self.assert_data(6, [{'a': 1, 'b': 1, 'c': 0, 'd': 1}], "data error")

        self.assert_data(8, [{'a': 0, 'b': 0, 'c': 1, 'd': 0, 'f': 0}], "data error")

        self.assert_data(10, [{'a': 0, 'b': 1, 'c': 0, 'd': 0}], "data error")

        self.assert_data(12, [{'a': 1, 'b': 1, 'c': 1, 'd': 1, 'e': 1, 'f': 0}], "data error")

        self.assert_data(14, [{'a': 0, 'b': 0, 'c': 0, 'd': 0, 'e': 0, 'f': 1}], "data error")

        self.assert_data(16, [{'a': 1, 'b': 0, 'c': 1, 'd': 0}], "data error")

        self.assert_data(18, [{'a': 0, 'b': 1, 'c': 0, 'd': 1}], "data error")

        self.assert_data(20, [{'a': 1, 'b': 'B'}], "data error")

        self.assert_data(22, [{'a': 36.2, 'b': 1, 'c': 'C'}], "data error")

        self.assert_data(24, [{'order_id': 2, 'uid': 1, 'goods_id': 1, 'amount': 0.6}], "data error")

        self.assert_data(26, [{'order_id': 2, 'uid': 1, 'goods_id': 1, 'amount': 0.6, 'status': 0}], "data error")

        self.assert_data(28, [{'a': 0, 'b': 1, 'c': 4, 'd': 'a X c', 'e': 'abc'}], "data error")

        self.assert_data(30, [{'a': 1, 'b': 0, 'c': 12, 'd': 'abc X ghi', 'e': 'def'}], "data error")


# -*- coding: utf-8 -*-
# 2023/5/26
# create by: snower

from syncany.calculaters.calculater import Calculater


class CurrentEnvVariableCalculater(Calculater):
    def __init__(self, *args):
        super(CurrentEnvVariableCalculater, self).__init__(*args)

        if self.name[22:] == "get_value":
            self.func = self.get_value
        elif self.name[22:] == "set_value":
            self.func = self.set_value
        else:
            self.func = lambda *args: None

    def calculate(self, *args):
        return self.func(*args)

    def get_value(self, key):
        from ..executor import Executor
        current_executor = Executor.current()
        if not current_executor:
            return None
        try:
            return current_executor.env_variables.get_value(key)
        except KeyError:
            return None

    def set_value(self, key, value):
        from ..executor import Executor
        current_executor = Executor.current()
        if current_executor:
            current_executor.env_variables[key] = value
        return value

    def is_realtime_calculater(self):
        return True



# -*- coding: utf-8 -*-
# 2023/6/1
# create by: snower

from syncany.calculaters.calculater import Calculater
from syncany.filters import IntFilter, FloatFilter
from ..errors import SyncanySqlExecutorException


class WindowAggregateCalculater(Calculater):
    def __init__(self, *args, **kwargs):
        super(WindowAggregateCalculater, self).__init__(*args, **kwargs)

        if self.name.endswith("::aggregate"):
            self.calculate = self.aggregate
        elif self.name.endswith("::order_aggregate"):
            self.calculate = self.order_aggregate

    def aggregate(self, state_value, data_value, context):
        raise SyncanySqlExecutorException("window caculate require order by")

    def order_aggregate(self, state_value, data_value, context):
        return context.current_index + 1

    def calculate(self, *args):
        if self.name.endswith("::aggregate"):
            return self.aggregate(*args)
        if self.name.endswith("::order_aggregate"):
            return self.order_aggregate(*args)
        return None


class WindowStateAggregateCalculater(WindowAggregateCalculater):
    def __init__(self, *args, **kwargs):
        super(WindowStateAggregateCalculater, self).__init__(*args, **kwargs)

        if self.name.endswith("::final_value"):
            self.calculate = self.final_value

    def final_value(self, state_value):
        return state_value

    def calculate(self, *args):
        if self.name.endswith("::final_value"):
            return self.final_value(*args)
        return super(WindowStateAggregateCalculater, self).calculate(*args)


class WindowAggregateRowNumberCalculater(WindowAggregateCalculater):
    def order_aggregate(self, state_value, data_value, context):
        return context.current_index + 1

    def get_final_filter(self):
        return IntFilter.default()


class WindowAggregateRankCalculater(WindowStateAggregateCalculater):
    def order_aggregate(self, state_value, data_value, context):
        order_value = context.order_value
        if state_value is None:
            return {"order_value": order_value, "rank": 1, "row_number": 1}
        state_value["row_number"] += 1
        if order_value != state_value["order_value"]:
            state_value["rank"] += 1
        return state_value

    def final_value(self, state_value):
        if state_value is None:
            return 1
        return state_value["rank"]

    def get_final_filter(self):
        return IntFilter.default()


class WindowAggregateDenseRankCalculater(WindowStateAggregateCalculater):
    def order_aggregate(self, state_value, data_value, context):
        order_value = context.order_value
        if state_value is None:
            return {"order_value": order_value, "rank": 1}
        if order_value != state_value["order_value"]:
            state_value["rank"] += 1
        return state_value

    def final_value(self, state_value):
        if state_value is None:
            return 1
        return state_value["rank"]

    def get_final_filter(self):
        return IntFilter.default()


class WindowAggregatePercentRankCalculater(WindowStateAggregateCalculater):
    def order_aggregate(self, state_value, data_value, context):
        order_value = context.order_value
        if state_value is None:
            return {"order_value": order_value, "rank": 1, "row_number": 1}
        state_value["row_number"] += 1
        if order_value != state_value["order_value"]:
            state_value["rank"] += 1
        return state_value

    def final_value(self, state_value):
        if state_value is None:
            return 1
        if state_value["row_number"] <= 1:
            return 0
        return state_value["rank"] - 1 / state_value["row_number"] - 1

    def get_final_filter(self):
        return FloatFilter.default()


class WindowAggregateFirstValueCalculater(WindowAggregateCalculater):
    def order_aggregate(self, state_value, data_value, context):
        valuer_id = context.partition_calculater.calculate_valuer.valuer_id
        for partition_calculater in context.datas[0][2]:
            if partition_calculater.calculate_valuer.valuer_id == valuer_id:
                return partition_calculater.value
        return None


class WindowAggregateLastValueCalculater(WindowAggregateCalculater):
    def order_aggregate(self, state_value, data_value, context):
        valuer_id = context.partition_calculater.calculate_valuer.valuer_id
        for partition_calculater in context.datas[-1][2]:
            if partition_calculater.calculate_valuer.valuer_id == valuer_id:
                return partition_calculater.value
        return None


class WindowAggregateNthValueCalculater(WindowAggregateCalculater):
    def order_aggregate(self, state_value, data_value, context):
        if isinstance(data_value, list):
            offset_index = int(data_value[1]) if len(data_value) >= 2 else 1
            if offset_index > 0:
                offset_index -= 1
            if (offset_index < 0 and -offset_index > len(context.datas)) or offset_index >= len(context.datas):
                return data_value[2] if len(data_value) >= 3 else None
            valuer_id = context.partition_calculater.calculate_valuer.valuer_id
            for partition_calculater in context.datas[offset_index][2]:
                if partition_calculater.calculate_valuer.valuer_id == valuer_id:
                    return partition_calculater.value[0]
            return None

        valuer_id = context.partition_calculater.calculate_valuer.valuer_id
        for partition_calculater in context.datas[0][2]:
            if partition_calculater.calculate_valuer.valuer_id == valuer_id:
                return partition_calculater.value
        return None


class WindowAggregateNtileCalculater(WindowAggregateCalculater):
    def order_aggregate(self, state_value, data_value, context):
        if isinstance(data_value, list):
            n = int(data_value[0]) if data_value else 1
        else:
            n = int(data_value)
        if n <= 0:
            return None
        return int(context.current_index / (len(context.datas) / n)) + 1


class WindowAggregateLagCalculater(WindowAggregateCalculater):
    def order_aggregate(self, state_value, data_value, context):
        if isinstance(data_value, list):
            offset = int(data_value[1]) if len(data_value) >= 2 else 1
            offset_index = context.current_index - offset
            if offset_index < 0 or offset_index >= len(context.datas):
                return data_value[2] if len(data_value) >= 3 else None
            valuer_id = context.partition_calculater.calculate_valuer.valuer_id
            for partition_calculater in context.datas[offset_index][2]:
                if partition_calculater.calculate_valuer.valuer_id == valuer_id:
                    return partition_calculater.value[0]
            return None

        offset_index = context.current_index - 1
        if offset_index < 0 or offset_index >= len(context.datas):
            return None
        valuer_id = context.partition_calculater.calculate_valuer.valuer_id
        for partition_calculater in context.datas[offset_index][2]:
            if partition_calculater.calculate_valuer.valuer_id == valuer_id:
                return partition_calculater.value
        return None


class WindowAggregateLeadCalculater(WindowAggregateCalculater):
    def order_aggregate(self, state_value, data_value, context):
        if isinstance(data_value, list):
            offset = int(data_value[1]) if len(data_value) >= 2 else 1
            offset_index = context.current_index + offset
            if offset_index < 0 or offset_index >= len(context.datas):
                return data_value[2] if len(data_value) >= 3 else None
            valuer_id = context.partition_calculater.calculate_valuer.valuer_id
            for partition_calculater in context.datas[offset_index][2]:
                if partition_calculater.calculate_valuer.valuer_id == valuer_id:
                    return partition_calculater.value[0]
            return None

        offset_index = context.current_index + 1
        if offset_index < 0 or offset_index >= len(context.datas):
            return None
        valuer_id = context.partition_calculater.calculate_valuer.valuer_id
        for partition_calculater in context.datas[offset_index][2]:
            if partition_calculater.calculate_valuer.valuer_id == valuer_id:
                return partition_calculater.value
        return None


class WindowAggregateCumeDistCalculater(WindowAggregateCalculater):
    def order_aggregate(self, state_value, data_value, context):
        if not context.datas:
            return 0
        return context.current_index + 1 / len(context.datas) + 1

    def get_final_filter(self):
        return FloatFilter.default()



# -*- coding: utf-8 -*-
# 2023/4/23
# create by: snower

from unittest import TestCase
from syncanysql import ScriptEngine


class ScriptEngineTestCase(TestCase):
    def get_test_data(self):
        return [
            {"site_id": 8, "site_name": "黄豆网", "site_amount": 17.04, "timeout_at": "16:00:00",
             "vip_timeout_at": "11:00:00"},
            {"site_id": 9, "site_name": "青菜网", "site_amount": 7.2, "timeout_at": "15:00:00",
             "vip_timeout_at": "11:00:00"},
            {"site_id": 8, "site_name": "去啥网", "site_amount": 10.41, "timeout_at": "16:00:00",
             "vip_timeout_at": "11:00:00"},
            {"site_id": 10, "site_name": "汽车网", "site_amount": 5.8, "timeout_at": "16:00:00",
             "vip_timeout_at": "11:00:00"},
            {"site_id": 9, "site_name": "火箭网", "site_amount": 4.5, "timeout_at": "15:00:00",
             "vip_timeout_at": "10:00:00"},
            {"site_id": 9, "site_name": "卫星网", "site_amount": 11.2, "timeout_at": "16:40:00",
             "vip_timeout_at": "11:20:00"},
        ]

    def get_stats_data(self):
        return [
            {'site_id': 8, 'site_name': '黄豆网', 'order_cnt': 2, 'total_amount': 27.45},
            {'site_id': 9, 'site_name': '青菜网', 'order_cnt': 3, 'total_amount': 22.9},
            {'site_id': 10, 'site_name': '汽车网', 'order_cnt': 1, 'total_amount': 5.8}
        ]

    def test_script_engine(self):
        with ScriptEngine() as engine:
            engine.push_memory_datas("test_data", self.get_test_data())
            self.assertEqual(engine.get_memory_datas("test_data"), self.get_test_data())

            engine.execute('''
                INSERT INTO `stats_data` SELECT site_id, site_name, count(*) order_cnt, sum(site_amount) total_amount 
                FROM test_data GROUP BY site_id;
            ''')
            self.assertEqual(engine.get_memory_datas("stats_data"), self.get_stats_data())
            self.assertEqual(engine.pop_memory_datas("stats_data"), self.get_stats_data())
            self.assertEqual(engine.get_memory_datas("stats_data"), [])

            engine.use("test_func", lambda: 1)
            engine.set_variable("test_var", 1)

            with engine.context() as context:
                context.use("test_func", lambda: 2)
                context.set_variable("test_var", 2)
                context.use("test_context_func", lambda: 2)
                context.set_variable("test_context_var", 2)

                context.execute('''
                    insert into `result_data` select test_func() as test_func, @test_var as test_var,
                     test_context_func() as test_context_func, @test_context_var as test_context_var;
                     
                     set global @aaa=1;
                ''')
                self.assertEqual(context.get_memory_datas("result_data"),
                                 [{"test_func": 2, "test_var": 2, "test_context_func": 2, "test_context_var": 2}])
                self.assertEqual(context.pop_memory_datas("result_data"),
                                 [{"test_func": 2, "test_var": 2, "test_context_func": 2, "test_context_var": 2}])
                self.assertEqual(context.get_memory_datas("result_data"), [])

            engine.execute('''
                insert into `result_data` select test_func() as test_func, @test_var as test_var;
                                
                insert into `result_aaa` select @aaa as aaa;
            ''')
            self.assertEqual(engine.get_memory_datas("result_data"),
                             [{"test_func": 1, "test_var": 1}])
            self.assertEqual(engine.pop_memory_datas("result_data"),
                             [{"test_func": 1, "test_var": 1}])
            self.assertEqual(engine.get_memory_datas("result_data"), [])

            self.assertEqual(engine.pop_memory_datas("result_aaa"), [{"aaa": 1}])




# -*- coding: utf-8 -*-
# 2025/3/25
# create by: snower

from .example import ExampleTestCase


class SubqueryExampleTestCase(ExampleTestCase):
    example_name = "subquery"

    def test_subquery(self):
        self.execute("subquery.sql")

        self.assert_data(1, [{'cnt': 4, 'latest_order_id': 6, 'total_amount': 27.6, 'uid': 2}], "data error")
        self.assert_data(5, [{'amount': 3,
                              'history_count': 5,
                              'history_exists': 1,
                              'order_id': 3,
                              'uid': 2}], "data error")
        self.assert_data(9, [{'goods_id': 1, 'goods_name': '青菜', 'order_count': 4, 'order_exists': 1},
                             {'goods_id': 2, 'goods_name': '白菜', 'order_count': 2, 'order_exists': 1},
                             {'goods_id': 3, 'goods_name': '萝卜', 'order_count': None, 'order_exists': 0}],
                         "data error")
        self.assert_data(13, [{'amount': 3,
                               'history_count': 5,
                               'history_exists': 1,
                               'order_id': 3,
                               'uid': 2}], "data error")
        self.assert_data(17, [{'goods_id': 1, 'goods_name': '青菜', 'order_count': 4, 'order_exists': 1},
                              {'goods_id': 2, 'goods_name': '白菜', 'order_count': 2, 'order_exists': 1},
                              {'goods_id': 3, 'goods_name': '萝卜', 'order_count': None, 'order_exists': 0}],
                         "data error")
        self.assert_data(21, [{'cnt': 1, 'order_id': 1, 'total_amount': 9.6},
                              {'cnt': 1, 'order_id': 2, 'total_amount': 7.6},
                              {'cnt': 1, 'order_id': 3, 'total_amount': 3}],
                         "data error")
        self.assert_data(23, [{'total_amount': 15.6, 'uid': 1}, {'total_amount': 27.6, 'uid': 2}],
                         "data error")
        self.assert_data(28, [{'amount': 3,
                               'has_history': 1,
                               'history_exists': 1,
                               'order_id': 3,
                               'uid': 2}], "data error")
        self.assert_data(36, [{'amount': 3,
                               'has_history': 1,
                               'history_exists': 1,
                               'order_id': 3,
                               'uid': 2}], "data error")
        self.assert_data(44, [{'amount': 3,
                               'has_history': 1,
                               'history_exists': 1,
                               'order_id': 3,
                               'uid': 2}], "data error")



# -*- coding: utf-8 -*-
# 2023/3/2
# create by: snower

import os
import types
from syncany.calculaters import CALCULATERS, Calculater, TypeFormatCalculater, TypingCalculater, MathematicalCalculater
from syncany.calculaters import register_calculater, find_calculater, CalculaterUnknownException
from syncany.calculaters.import_calculater import ImportCalculater
from .env_variable_calculater import CurrentEnvVariableCalculater
from .query_tasker_calculater import ExecuteQueryTaskerCalculater
from .row_calculater import *
from .mysql_calculater import MysqlCalculater, register_mysql_func
from .generate_calculater import *
from .aggregate_calculater import *
from .window_calculater import *

SQL_CALCULATERS = {
    "current_env_variable": CurrentEnvVariableCalculater,
    "row_index": RowIndexCalculater,
    "row_last": RowLastCalculater,
    "mysql": MysqlCalculater,
    "aggregate_key": AggregateKeyCalculater,
    "aggregate_count": AggregateCountCalculater,
    "aggregate_distinct_count": AggregateDistinctCountCalculater,
    "aggregate_sum": AggregateSumCalculater,
    "aggregate_max": AggregateMaxCalculater,
    "aggregate_min": AggregateMinCalculater,
    "aggregate_avg": AggregateAvgCalculater,
    "yield_array": GenerateYieldArrayCalculater,
    "group_concat": AggregateGroupConcatCalculater,
    "group_array": AggregateGroupArrayCalculater,
    "group_uniq_array": AggregateGroupUniqArrayCalculater,
    "group_bit_and": AggregateGroupBitAndCalculater,
    "group_bit_or": AggregateGroupBitOrCalculater,
    "group_bit_xor": AggregateGroupBitXorCalculater,
    "json_arrayagg": AggregateJsonArrayaggCalculater,
    "json_objectagg": AggregateJsonObjectaggCalculater,
    "aggregate_any_value": AggregateAnyValueCalculater,
    "row_number": WindowAggregateRowNumberCalculater,
    "rank": WindowAggregateRankCalculater,
    "dense_rank": WindowAggregateDenseRankCalculater,
    "percent_rank": WindowAggregatePercentRankCalculater,
    "cume_dist": WindowAggregateCumeDistCalculater,
    "first_value": WindowAggregateFirstValueCalculater,
    "last_value": WindowAggregateLastValueCalculater,
    "nth_value": WindowAggregateNthValueCalculater,
    "ntile": WindowAggregateNtileCalculater,
    "lag": WindowAggregateLagCalculater,
    "lead": WindowAggregateLeadCalculater,
    "execute_query_tasker": ExecuteQueryTaskerCalculater,
}
CALCULATERS.update(SQL_CALCULATERS)
if not os.environ.get("SYNCANY_PYEVAL_DISABLED"):
    from .pyeval_calculater import PyEvalCalculater, register_pyeval_module
    CALCULATERS["pyeval"] = PyEvalCalculater
    CALCULATERS["pyevalt"] = PyEvalCalculater


def is_mysql_func(name):
    if MysqlCalculater.funcs is None:
        MysqlCalculater.find_func(name)
    return name in MysqlCalculater.funcs


def find_generate_calculater(name):
    calculater = find_calculater(name)
    if issubclass(calculater, ImportCalculater):
        try:
            import_calculater = ImportCalculater(name)
            if isinstance(import_calculater.module_or_func, (types.FunctionType, types.LambdaType)):
                if import_calculater.module_or_func.__code__.co_flags & 0x20 != 0:
                    return calculater
        except:
            pass
        raise CalculaterUnknownException("%s is unknown generate calculater" % name)
    if not issubclass(calculater, GenerateCalculater):
        raise CalculaterUnknownException("%s is unknown generate calculater" % name)
    return calculater


def find_aggregate_calculater(name):
    calculater = find_calculater(name)
    if not issubclass(calculater, AggregateCalculater):
        raise CalculaterUnknownException("%s is unknown aggregate calculater" % name)
    return calculater


def find_window_aggregate_calculater(name):
    calculater = find_calculater(name)
    if not issubclass(calculater, WindowAggregateCalculater):
        raise CalculaterUnknownException("%s is unknown window aggregate calculater" % name)
    return calculater


# -*- coding: utf-8 -*-
# 2023/2/25
# create by: snower

import os


class UseCommandTasker(object):
    def __init__(self, config):
        self.config = config

    def start(self, name, executor, session_config, manager, arguments):
        use_info = [s.strip() for s in self.config["use"].split(" as ")]
        if not use_info[0] or isinstance(use_info[0], (bool, int, float, list, tuple, set, dict)):
            return []
        if isinstance(use_info[0], str):
            __import__(use_info[0], globals(), locals(), [use_info[0].rpartition(".")[-1]])
        if len(use_info) >= 2:
            if os.path.exists(use_info[0]):
                session_config.set("sources." + use_info[1], use_info[0])
            else:
                session_config.set("imports." + use_info[1], use_info[0])
        else:
            if os.path.exists(use_info[0]):
                session_config.set("sources." + use_info[0].split(".")[-1], use_info[0])
            else:
                session_config.set("imports." + use_info[0].split(".")[-1], use_info[0])
        return []

    def run(self, executor, session_config, manager):
        pass

    def terminate(self):
        pass



# -*- coding: utf-8 -*-
# 2023/4/21
# create by: snower

from .example import ExampleTestCase


class MemoryTemporaryStorageExampleTestCase(ExampleTestCase):
    example_name = "memory_temporary_storage"

    def test_memory_temporary_storage(self):
        self.execute("memory_temporary_storage.sql")

        self.assert_data(6, [
            {'Id': 'aa7e941b-d399-4bec-0ba4-08d8dd2f9239', 'Email': 'bm6U11zDIspdNW1iQiVZdHX8uqOWZe0cers9BZEcCrE=',
             'FirstName': 'John', 'LastName': 'Doe'}], "data error")
        self.assert_data(8, [{'Id': 'aa7e941b-d399-4bec-0ba4-08d8dd2f9239', 'Name': 'LiMei'}], "data error")
        self.assert_data(10, [
            {'Id': 'aa7e941b-d399-4bec-0ba4-08d8dd2f9239', 'Email': 'bm6U11zDIspdNW1iQiVZdHX8uqOWZe0cers9BZEcCrE=',
             'Name': 'LiMei', 'FirstName': 'John', 'LastName': 'Doe'}], "data error")



# -*- coding: utf-8 -*-
# 2023/2/8
# create by: snower

from syncany.errors import SyncanyException

class SyncanySqlException(SyncanyException):
    pass

class SyncanySqlCompileException(SyncanySqlException):
    pass

class SyncanySqlExecutorException(SyncanySqlException):
    pass


# -*- coding: utf-8 -*-
# 2023/4/21
# create by: snower

import datetime
import uuid
from bson.objectid import ObjectId
from .example import ExampleTestCase


class TypeAnnotationExampleTestCase(ExampleTestCase):
    example_name = "type_annotation"

    def test_type_annotation(self):
        self.execute("type_annotation.sql")

        self.assert_data(5, [{'t': 'str', 'v': '3243243'}], "data error")
        self.assert_data(7, [{'t': 'str', 'v': '12.90'}], "data error")
        self.assert_value(10, 't', 'datetime', "data error")
        self.assert_value(10, 'v', lambda value: isinstance(value, datetime.datetime), "data error")
        self.assert_value(12, 'orderAt', lambda value: isinstance(value, datetime.datetime), "data error")
        self.assert_value(12, 't', 'str', "data error")
        self.assert_value(12, 'v', '2023-02-23', "data error")
        self.assert_value(16, 'orderAt', lambda value: isinstance(value, datetime.datetime), "data error")
        self.assert_value(16, 't', 'int', "data error")
        self.assert_value(16, 'v', lambda value: isinstance(value, int), "data error")
        self.assert_data(21, [{'t': 'objectid', 'v': ObjectId('6422a08fb4055348da72633b')},
                              {'t': 'objectid', 'v': ObjectId('6422a0a4b4055348da72633c')}], "data error")
        self.assert_data(26, [{'t': 'uuid', 'v': uuid.UUID('da984ae6-cd3f-11ed-af1b-eb91b1b4fa12')}], "data error")
        self.assert_data(28, [{'payId': uuid.UUID('da984ae6-cd3f-11ed-af1b-eb91b1b4fa12'), 't': 'int',
                               'v': 290562451387917676601713545966572075538}], "data error")
        self.assert_data(33, [{'t': 'bool', 'v': True}], "data error")
        self.assert_data(34, [{'v1': False, 'v2': True, 'v3': False, 'v4': True, 'v5': None}], "data error")

    def test_type_declaration_cast(self):
        self.execute("type_declaration_cast.sql")

        self.assert_data(5, [{'t': 'str', 'v': '3243243'}], "data error")
        self.assert_data(7, [{'t': 'str', 'v': '12.90'}], "data error")
        self.assert_value(10, 't', 'datetime', "data error")
        self.assert_value(10, 'v', lambda value: isinstance(value, datetime.datetime), "data error")
        self.assert_value(12, 'orderAt', lambda value: isinstance(value, datetime.datetime), "data error")
        self.assert_value(12, 't', 'str', "data error")
        self.assert_value(12, 'v', '2023-02-23', "data error")
        self.assert_value(16, 'orderAt', lambda value: isinstance(value, datetime.datetime), "data error")
        self.assert_value(16, 't', 'int', "data error")
        self.assert_value(16, 'v', lambda value: isinstance(value, int), "data error")
        self.assert_data(21, [{'t': 'objectid', 'v': ObjectId('6422a08fb4055348da72633b')},
                              {'t': 'objectid', 'v': ObjectId('6422a0a4b4055348da72633c')}], "data error")
        self.assert_data(26, [{'t': 'uuid', 'v': uuid.UUID('da984ae6-cd3f-11ed-af1b-eb91b1b4fa12')}], "data error")
        self.assert_data(28, [{'payId': uuid.UUID('da984ae6-cd3f-11ed-af1b-eb91b1b4fa12'), 't': 'int',
                               'v': 290562451387917676601713545966572075538}], "data error")
        self.assert_data(33, [{'t': 'bool', 'v': True}], "data error")
        self.assert_data(34, [{'v1': False, 'v2': True, 'v3': False, 'v4': True, 'v5': None}], "data error")



# -*- coding: utf-8 -*-
# 2023/2/8
# create by: snower

import os
from .errors import SyncanySqlCompileException


class SqlSegment(object):
    def __init__(self, sql, lineno=None):
        self.sql = sql
        self.lineno = lineno

    def __str__(self):
        return self.sql


class SqlParser(object):
    ESCAPE_CHARS = {'\a', '\b', '\f', '\n', '\r', '\t', '\v', '\\', '\'', '"', '\0'}

    def __init__(self, sql):
        self.sql = sql
        self.index = 0
        self.lineno = 1
        self.len = len(sql)

    def next(self):
        if self.index >= len(self.sql):
            return
        if self.sql[self.index] == '\n':
            self.lineno += 1
        self.index += 1

    def reset(self):
        self.index = 0
        self.lineno = 1

    def skip_escape(self, c):
        start_index = self.index
        self.next()
        while self.index < self.len:
            if self.sql[self.index] != c:
                self.next()
                continue
            backslash = self.index - 1
            while backslash >= 0:
                if self.sql[backslash] != '\\':
                    break
                backslash -= 1
            if (self.index - backslash + 1) % 2 != 0:
                self.next()
                continue
            self.next()
            return start_index, self.index, self.sql[start_index: self.index]
        self.next()
        raise EOFError(self.sql[start_index:])

    def read_util(self, cs, escape_chars=('"', "'")):
        start_index = self.index
        while self.index < self.len:
            if self.sql[self.index] in escape_chars:
                self.skip_escape(self.sql[self.index])
                continue
            if self.sql[self.index: self.index + len(cs)] != cs:
                self.next()
                continue
            return start_index, self.index + len(cs) - 1, self.sql[start_index: self.index + len(cs) - 1]
        raise EOFError(self.sql[start_index:])

    def read_segment(self, c='`'):
        while self.index < self.len:
            self.read_util(c)
            self.next()
            start_index, end_index, segment = self.read_util(c)
            self.next()
            return start_index, end_index, segment
        raise EOFError()

    def read_util_complete(self):
        while self.index < self.len:
            return self.read_util(';', ('`', '"', "'"))

    def split(self):
        segments = []
        lineno, start_index = 0, None
        while self.index < self.len:
            if self.sql[self.index] in ('`', '"', "'"):
                self.skip_escape(self.sql[self.index])
                continue
            if self.sql[self.index] == "#":
                try:
                    self.read_util('\n')
                except EOFError:
                    pass
                self.next()
                continue
            segment = self.sql[self.index: self.index + 2]
            if segment == "--":
                try:
                    self.read_util('\n')
                except EOFError:
                    pass
                self.next()
                continue
            if segment == "/*":
                self.read_util("*/")
                self.next()
                continue
            if self.sql[self.index] == ';':
                if start_index is None:
                    self.next()
                    continue
                if start_index < self.index:
                    segments.append(SqlSegment(self.sql[start_index: self.index], lineno))
                start_index = None
                self.next()
                continue
            if start_index is None and self.sql[self.index].isalpha():
                lineno, start_index = self.lineno, self.index
            self.next()
        if start_index is not None and start_index < self.index:
            segments.append(SqlSegment(self.sql[start_index: self.index], lineno))
        return segments


class FileParser(object):
    def __init__(self, filename):
        self.filename = filename

    def load(self):
        with open(self.filename, "r+", encoding=os.environ.get("SYNCANYENCODING", "utf-8")) as fp:
            content = fp.read()
        if isinstance(self.filename, str) and self.filename[-5:].lower() == ".prql":
            try:
                import prql_python
                content = prql_python.compile(content, options=prql_python.CompileOptions(
                    signature_comment=False, target="sql.mysql"))
            except ImportError:
                raise SyncanySqlCompileException("prql-python>=0.11.1 is required")
        sql_parser = SqlParser(content)
        sqls = sql_parser.split()
        return sqls



# -*- coding: utf-8 -*-
# 2023/2/27
# create by: snower

import copy


class ShowCommandTasker(object):
    def __init__(self, config):
        self.config = config

    def start(self, name, executor, session_config, manager, arguments):
        if self.config.get("key").lower() == "databases":
            self.show_databases(session_config)
        elif self.config.get("key").lower() == "imports":
            self.show_imports(session_config)
        return []

    def run(self, executor, session_config, manager):
        pass

    def terminate(self):
        pass

    def show_databases(self, session_config):
        databases = copy.deepcopy(session_config.get().get("databases", []))
        fields, datas = ["name", "driver", "params", "virtual_views"], []
        for database in databases:
            virtual_views = database.pop("virtual_views", None)
            datas.append({
                "name": database.pop("name"),
                "driver": database.pop("driver"),
                "params": database,
                "virtual_views": virtual_views if virtual_views else "",
            })
        self.print(fields, datas)

    def show_imports(self, session_config):
        imports = copy.deepcopy(session_config.get().get("imports", {}))
        fields, datas = ["name", "alias"], []
        for alias, name in imports.items():
            datas.append({
                "name": name,
                "alias": alias
            })
        self.print(fields, datas)

    def print(self, fields, datas):
        try:
            import rich
            from rich.table import Table
        except ImportError:
            print("\t".join(fields))
            for data in datas:
                print("\t".join([str(data[field]) for field in fields]))
            return

        table = Table(show_header=True, collapse_padding=True, expand=True, highlight=True)
        for field in fields:
            table.add_column(field)
        for data in datas:
            table.add_row(*(str(data[field]) for field in fields))
        rich.print(table)


# -*- coding: utf-8 -*-
# 2023/2/13
# create by: snower

import datetime
import os
import sys
import re
import threading
from collections import deque
from syncany.filters import StringFilter
from syncany.calculaters import find_calculater
from syncany.database import find_database
from syncany.taskers.config import ConfigReader, register_reader
from syncany.taskers.core import CoreTasker
from .errors import SyncanySqlCompileException
from .utils import SequenceTypes, parse_value
from .compiler import Compiler

ENV_VARIABLE_RE = re.compile(r"(\$\{[@\w]+?(:.*?)?\})", re.DOTALL | re.M)
RAW_SQL_RE = re.compile(r"(\/\*\s*raw\(([\w\.]+?)\)\s*(\*\/\s*\()?\s*(.*?)\s*(\)\s*\/\*)?\s*endraw\s*\*\/)", re.DOTALL | re.M)
FUNC_RE = re.compile(r"^(\w+?)\(((.+),?)*\)$", re.DOTALL)


@register_reader("context")
class ContextConfigReader(ConfigReader):
    def read(self):
        executor = Executor.current()
        if not executor:
            return "dict", {}
        names = self.name[10:].split("/")
        if len(names) >= 2 and names[0] == "session" and names[1] == "config":
            if len(names) >= 3:
                return "dict", {
                    names[2]: executor.session_config.get().get(names[2], CoreTasker.DEFAULT_CONFIG.get(names[2]))
                }
            return "dict", executor.session_config.get_tasker_extend_config()
        return "dict", {}


class EnvVariables(dict):
    def __init__(self, parent=None, *args, **kwargs):
        super(EnvVariables, self).__init__(*args, **kwargs)

        self.parent = parent

    @classmethod
    def build_global(cls, session_config):
        env_variables = cls()
        env_variables.update({key.lower(): value for key, value in os.environ.items()})
        env_variables["home"] = os.path.expanduser('~')
        env_variables["syncany_home"] = session_config.get_home()
        env_variables["cwd"] = os.getcwd()
        env_variables["platform"] = sys.platform
        for arg in sys.argv:
            if arg[:2] != "--":
                continue
            arg_info = arg[2:].split("=")
            arg_value = arg_info[1] if len(arg_info) >= 2 else True
            if isinstance(arg_value, str) and arg_value and arg_value[0] in ('"', "'"):
                arg_value = arg_value[1:-1]
            env_variables[arg_info[0]] = arg_value
        return env_variables

    def get_value(self, key):
        if key in self:
            return self[key]
        if key and key[0] == "@" and key[1:] in self:
            return self[key[1:]]
        if self.parent is None:
            raise KeyError("%s unknown" % key)
        return self.parent.get_value(key)

    def get(self, key, default=None):
        if key in self:
            return self[key]
        if key and key[0] == "@" and key[1:] in self:
            return self[key[1:]]
        if self.parent is None:
            return default
        return self.parent.get(key, default)


class Executor(object):
    _thread_local = threading.local()
    global_env_variables = None
    tasker_index = 0

    @classmethod
    def current(cls):
        return cls._thread_local.current_executor

    def __init__(self, manager, session_config, parent_executor=None):
        self.manager = manager
        self.session_config = session_config
        self.parent_executor = parent_executor
        self.runners = deque()
        self.tasker = None
        if self.global_env_variables is None:
            self.__class__.global_env_variables = EnvVariables.build_global(session_config)
        self.env_variables = EnvVariables(parent_executor.env_variables if parent_executor
                                          else self.global_env_variables)

    def compile_variable(self, sql):
        variables = ENV_VARIABLE_RE.findall(sql)
        for variable, default_value in variables:
            variable_name = variable[2:-1].split(":")[0]
            try:
                variable_value = self.env_variables.get_value(variable_name.lower())
            except KeyError:
                try:
                    groups = FUNC_RE.match(default_value[1:].strip()).groups()
                    calculater_args = []
                    if groups[1]:
                        for arg in groups[1].split(","):
                            calculater_args.append(parse_value(arg))
                    calculater = find_calculater(groups[0].split("__")[0]).instance(groups[0].replace("__", "::"))
                    variable_value = calculater.calculate(*tuple(calculater_args))
                    variable_value = StringFilter().filter(variable_value)
                except Exception as e:
                    variable_value = default_value[1:]

            def format_variable_value(value):
                if value is True:
                    return "true"
                if value is False:
                    return "false"
                if value is None:
                    return "null"
                if isinstance(value, datetime.date):
                    if isinstance(value, datetime.datetime):
                        return value.strftime(self.session_config.get().get("datetime_format", "%Y-%m-%d %H:%M:%S"))
                    return value.strftime(self.session_config.get().get("date_format", "%Y-%m-%d"))
                if isinstance(value, datetime.time):
                    return value.strftime(self.session_config.get().get("time_format", "%H:%M:%S"))
                if isinstance(value, SequenceTypes):
                    return "(" + ",".join([format_variable_value(v) for v in value]) + ")"
                return str(value)
            sql = sql.replace(variable, format_variable_value(variable_value))
        return sql

    def run(self, name, sqls):
        self._thread_local.current_executor = self
        for sql in sqls:
            lineno = sql.lineno
            sql = self.compile_variable(str(sql))
            raw_sqls = RAW_SQL_RE.findall(sql)
            for raw, raw_name, _, raw_sql, _ in raw_sqls:
                raw_name_info = raw_name.split(".")
                if len(raw_name_info) != 2:
                    raise SyncanySqlCompileException("raw sql name error: %s", raw)
                raw_sql = "set @config.databases." + raw_name_info[0] + ".virtual_views." \
                          + raw_name_info[1] + "='''\n" + raw_sql.strip() + "\n'''"
                self.compile(name + "(" + str(lineno) + ")#set_virtual_view", raw_sql)
                sql = sql.replace(raw, raw_name)
            self.compile(name + "(" + str(lineno) + ")", sql)

    def compile(self, name, sql):
        self._thread_local.current_executor = self
        compiler = Compiler(self.session_config, self.env_variables, name)
        arguments = {"@verbose": self.env_variables.get("@verbose", False), "@timeout": self.env_variables.get("@timeout", 0),
                     "@limit": self.env_variables.get("@limit", 0), "@batch": self.env_variables.get("@batch", 0),
                     "@streaming": self.env_variables.get("@streaming", False), "@recovery": self.env_variables.get("@recovery", False),
                     "@join_batch": self.env_variables.get("@join_batch", 10000), "@insert_batch": self.env_variables.get("@insert_batch", 0),
                     "@use_input": self.env_variables.get("@use_input", None), "@use_output": self.env_variables.get("@use_output", None),
                     "@use_output_type": self.env_variables.get("@use_output_type", None), "@primary_order": False}
        tasker = compiler.compile(sql, arguments)
        self.runners.extend(tasker.start(name, self, self.session_config, self.manager, arguments))

    def execute(self):
        self._thread_local.current_executor = self
        while self.runners:
            self.tasker = self.runners.popleft()
            try:
                self.tasker.run(self, self.session_config, self.manager)
            finally:
                self.tasker = None

    def terminate(self):
        if not self.tasker:
            return
        self.tasker.terminate()

    def add_runner(self, tasker):
        self.runners.append(tasker)

    def clear_temporary_memory_collection(self, names):
        try:
            database_config = dict(**[database for database in self.session_config.get()["databases"]
                                      if database["name"] == "--"][0])
        except (TypeError, KeyError, IndexError):
            return
        database_cls = find_database(database_config.pop("driver"))
        if not database_cls:
            return
        database = database_cls(self.manager.database_manager, database_config).build()
        try:
            for name in names:
                if self.runners and database.is_streaming(name):
                    continue
                delete = database.delete(name, ["id"])
                delete.commit()
        finally:
            database.close()

    def __enter__(self):
        self._thread_local.current_executor = self
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._thread_local.current_executor = self.parent_executor

    def distribute_tasker_index(self):
        if self.parent_executor is not None:
            return self.parent_executor.distribute_tasker_index()
        self.tasker_index += 1
        return self.tasker_index

    def get_tasker_index(self):
        if self.parent_executor is not None:
            return self.parent_executor.get_tasker_index()
        return self.tasker_index



# -*- coding: utf-8 -*-
# 2023/2/19
# create by: snower

from decimal import Decimal
import datetime
import string

NumberTypes = (int, float)
NumberDecimalTypes = (int, float, Decimal)
NumberStringTypes = (int, float, str)
SequenceTypes = (tuple, list)


def parse_value(value):
    value = value.strip()
    if not value:
        return ""
    if len(value) >= 2 and value[0] in ("'", '"') and value[-1] in ("'", '"'):
        return value[1:-1]
    if value.isdigit() or (value[0] == "-" and value[1:].isdigit()):
        return int(value)
    if value.lower() == "true":
        return True
    if value.lower() == "false":
        return False
    if value.lower() == "null":
        return None
    value_info = value.split(".")
    if len(value_info) == 2 and (value_info[0].isdigit() or (value[0][0] == "-" and value[0][1:].isdigit())) \
            and value_info[1].isdigit():
        return float(value)
    return value

def parse_number(x, is_float=False):
    index, dot_index = -1, -1
    for i in range(len(x)):
        if x[i] in string.digits:
            index = i
            continue
        if index >= 0 and is_float and dot_index < 0 and x[i] == ".":
            dot_index = i
            continue
        break
    if index < 0:
        return float(x) if is_float else int(x)
    return float(x[:index + 1]) if dot_index > 0 else int(x[:index + 1])

def ensure_int(x):
    if x.__class__ is int:
        return x
    if x is None:
        raise ValueError('value is None')
    if not x:
        return 0
    if x is True:
        return 1
    if isinstance(x, NumberTypes):
        return int(x)
    if isinstance(x, datetime.date):
        if isinstance(x, datetime.datetime):
            return int(x.strftime("%Y%m%d%H%M%S"))
        return int(x.strftime("%Y%m%d"))
    if isinstance(x, datetime.time):
        return int(x.strftime("%H%M%S"))
    if isinstance(x, str):
        try:
            return int(x)
        except:
            return parse_number(x, False)
    return int(x)

def ensure_float(x):
    if x.__class__ is float:
        return x
    if x is None:
        raise ValueError('value is None')
    if not x:
        return 0
    if x is True:
        return 1
    if isinstance(x, NumberTypes):
        return float(x)
    if isinstance(x, datetime.date):
        if isinstance(x, datetime.datetime):
            return float(x.strftime("%Y%m%d%H%M%S")) + x.microsecond / 1000
        return float(x.strftime("%Y%m%d"))
    if isinstance(x, datetime.time):
        return float(x.strftime("%H%M%S")) + x.microsecond / 1000
    if isinstance(x, str):
        try:
            return float(x)
        except:
            return parse_number(x, True)
    return float(x)

def ensure_number(x):
    if x.__class__ is int or x.__class__ is float:
        return x
    if isinstance(x, NumberDecimalTypes):
        return x
    if x is None:
        raise ValueError('value is None')
    if not x:
        return 0
    if x is True:
        return 1
    if isinstance(x, str):
        if "." in x:
            try:
                return float(x)
            except:
                return ensure_float(x)
        try:
            return int(x)
        except:
            return ensure_int(x)
    if isinstance(x, datetime.date):
        if isinstance(x, datetime.datetime):
            return int(x.strftime("%Y%m%d%H%M%S"))
        return int(x.strftime("%Y%m%d"))
    if isinstance(x, datetime.time):
        return int(x.strftime("%H%M%S"))
    return ensure_int(x)

def ensure_str(x):
    if x.__class__ is str:
        return x
    if x is None:
        raise ValueError('value is None')
    if not x:
        return '0'
    if x is True:
        return '1'
    if isinstance(x, str):
        return str(x)
    if isinstance(x, bytes):
        return x.decode("utf-8")
    if isinstance(x, datetime.date):
        if isinstance(x, datetime.datetime):
            return x.strftime("%Y-%m-%d %H:%M:%S")
        return x.strftime("%Y-%m-%d")
    if isinstance(x, datetime.time):
        return x.strftime("%H:%M:%S")
    return str(x)



# -*- coding: utf-8 -*-
# 2023/4/21
# create by: snower

from .example import ExampleTestCase


class GetValueExampleTestCase(ExampleTestCase):
    example_name = "get_value"

    def test_get_value(self):
        self.execute("get_value.sql")

        self.assert_data(4, [
            {'orderId': 3243243, 'username': 'snower', 'payId': '23232323232323', 'refundId': ['111111']}],
                         "data error")
        self.assert_data(6, [
            {'orderId': 3243243, 'username': 'snower', 'payId': '23232323232323', 'refundId': ['111111']}],
                         "data error")
        self.assert_data(8, [{'mobile1': '12345678911', 'mobile2': '12345678911'}], "data error")
        self.assert_data(10, [{'payTypes1': 1, 'payTypes2': 1, 'payTypes3': 1}], "data error")
        self.assert_data(12, [{'payTypes1': [1, 3], 'payTypes2': [1, 3], 'payTypes3': [1, 3]}], "data error")
        self.assert_data(14, [{'payTypes1': [4, 3], 'payTypes2': [4, 3], 'payTypes3': [4, 3]}], "data error")
        self.assert_data(16, [{'channel1': 'weixin', 'channel2': 'weixin'}], "data error")



# -*- coding: utf-8 -*-
# 2023/3/2
# create by: snower

import types
import traceback
from syncany.logger import get_logger
from syncany.calculaters.calculater import Calculater
from syncany.calculaters.import_calculater import parse_final_filter


class MysqlCalculater(Calculater):
    funcs = None

    @classmethod
    def find_func(cls, name):
        if cls.funcs is None:
            from . import mysql_funcs
            cls.funcs = mysql_funcs.funcs
        return cls.funcs.get(name)

    def __init__(self, *args, **kwargs):
        super(MysqlCalculater, self).__init__(*args, **kwargs)

        func = self.find_func(self.name[7:])
        if func is not None and isinstance(func, types.FunctionType):
            try:
                argcount, varnames_count = func.__code__.co_argcount, len(func.__code__.co_varnames)
                if argcount == varnames_count:
                    if argcount == 0:
                        self.calculate = self.calculate0
                    elif argcount == 1:
                        self.calculate = self.calculate1
                    elif argcount == 2:
                        self.calculate = self.calculate2
                    elif argcount == 3:
                        self.calculate = self.calculate3
                    elif argcount == 4:
                        self.calculate = self.calculate4
            except Exception as e:
                pass
        self.func  = func

    def calculate(self, *args):
        try:
            return self.func(*args)
        except Exception as e:
            if isinstance(e, (ValueError, KeyError)):
                return None
            get_logger().warning("mysql calculater execute %s(%s) error: %s\n%s", self.name[7:], args, e,
                                 traceback.format_exc())
            return None

    def calculate0(self):
        try:
            return self.func()
        except Exception as e:
            if isinstance(e, (ValueError, KeyError)):
                return None
            get_logger().warning("mysql calculater execute %s(%s) error: %s\n%s", self.name[7:], tuple(), e,
                                 traceback.format_exc())
            return None

    def calculate1(self, args0=None):
        try:
            return self.func(args0)
        except Exception as e:
            if isinstance(e, (ValueError, KeyError)):
                return None
            get_logger().warning("mysql calculater execute %s(%s) error: %s\n%s", self.name[7:], (args0,), e,
                                 traceback.format_exc())
            return None

    def calculate2(self, args0=None, args1=None):
        try:
            return self.func(args0, args1)
        except Exception as e:
            if isinstance(e, (ValueError, KeyError)):
                return None
            get_logger().warning("mysql calculater execute %s(%s) error: %s\n%s", self.name[7:], (args0, args1), e,
                                 traceback.format_exc())
            return None

    def calculate3(self, args0=None, args1=None, args2=None):
        try:
            return self.func(args0, args1, args2)
        except Exception as e:
            if isinstance(e, (ValueError, KeyError)):
                return None
            get_logger().warning("mysql calculater execute %s(%s) error: %s\n%s", self.name[7:], (args0, args1, args2), e,
                                 traceback.format_exc())
            return None

    def calculate4(self, args0=None, args1=None, args2=None, args3=None):
        try:
            return self.func(args0, args1, args2, args3)
        except Exception as e:
            if isinstance(e, (ValueError, KeyError)):
                return None
            get_logger().warning("mysql calculater execute %s(%s) error: %s\n%s", self.name[7:], (args0, args1, args2, args3), e,
                                 traceback.format_exc())
            return None

    def get_final_filter(self):
        if hasattr(self.func, "get_final_filter"):
            return self.func.get_final_filter()
        return parse_final_filter(self.func)

    def is_realtime_calculater(self):
        if self.name[7:] in {"currenttimestamp", "curdate", "currentdate", "curtime", "currenttime",
                             "sysdate", "unix_timestamp", "utc_date", "utc_time", "utc_timestamp"}:
            return True
        return False


def register_mysql_func(name, func):
    if not isinstance(func, types.FunctionType):
        raise TypeError("func must be a function")
    if MysqlCalculater.funcs is None:
        from . import mysql_funcs
        MysqlCalculater.funcs = mysql_funcs.funcs
    MysqlCalculater.funcs[name] = func


# -*- coding: utf-8 -*-
# 2023/2/14
# create by: snower

import copy
import datetime
from syncany.logger import get_verbose_logger


class ExplainTasker(object):
    def __init__(self, tasker, sql=None):
        self.config = None
        self.tasker = tasker
        self.sql = sql

    def start(self, name, executor, session_config, manager, arguments):
        self.config = copy.deepcopy(self.tasker.config)
        if self.sql:
            get_verbose_logger()("%s tasker %s execute sql:\n%s" % (datetime.datetime.now(), self.config["name"], self.sql))
            print()
        get_verbose_logger()("%s tasker %s compiled config:" % (datetime.datetime.now(), self.config["name"]))
        get_verbose_logger()(self.config)
        print()

        for key in list(arguments.keys()):
            if key.endswith("@verbose"):
                arguments[key] = True
        self.tasker.start(name, executor, session_config, manager, arguments)
        return [self]

    def run(self, executor, session_config, manager):
        try:
            return self.tasker.run(executor, session_config, manager)
        finally:
            self.config, self.tasker = None, None

    def terminate(self):
        if not self.tasker:
            return
        self.tasker.terminate()
        self.tasker = None



# -*- coding: utf-8 -*-
# 2025/4/3
# create by: snower

from .example import ExampleTestCase


class PyEvalExampleTestCase(ExampleTestCase):
    example_name = "pyeval"

    def test_pyeval(self):
        self.execute("pyeval.sql")

        self.assert_data(2, [{'a': 3, 'b': '0,1,2,3', 'c': 3, 'd': '123'}],
                         "data error")

        self.assert_data(6, [{'a': 1}], "data error")

        self.assert_data(8, [{'v': 0}, {'v': 2}], "data error")


# -*- coding: utf-8 -*-
# 2023/4/21
# create by: snower

import sys
import os
import datetime
from .example import ExampleTestCase


class ImportPythonExampleTestCase(ExampleTestCase):
    example_name = "import_python"

    def test_import_python(self):
        self.execute("import_python.sql")

        self.assert_value(8, "UTILS$HELLO()", 'hello world!', "data error")

        self.assert_value(10, "UTILS$ADD_NUMBER(1, 2)", 3, "data error")
        self.assert_value(10, "UTILS$SUM_ARRAY((1, 2, 3))", 6, "data error")

        self.assert_value(12, "PARSING$PARSE('2023-02-10 10:33:22')",
                          lambda value: isinstance(value, datetime.datetime), "data error")

        self.assert_value(14, "SYS$VERSION()", sys.version, "data error")
        self.assert_value(14, "OS$GETCWD()", lambda value: os.getcwd() in value, "data error")

        self.assert_value(16, "PYTHON_DATETIME$DATETIME$NOW()",
                          lambda value: isinstance(value, datetime.datetime), "data error")



# -*- coding: utf-8 -*-
# 2023/3/2
# create by: snower

from . import number_funcs
from . import string_funcs
from . import datetime_funcs
from . import logical_funcs
from . import json_funcs
from . import regexp_funcs

funcs = {}
funcs.update(number_funcs.funcs)
funcs.update(string_funcs.funcs)
funcs.update(datetime_funcs.funcs)
funcs.update(logical_funcs.funcs)
funcs.update(json_funcs.funcs)
funcs.update(regexp_funcs.funcs)


# -*- coding: utf-8 -*-
# 2025/3/24
# create by: snower

import copy
from syncany.taskers.tasker import _thread_local
from syncany.calculaters.calculater import LoaderCalculater


class ExecuteQueryTaskerCalculater(LoaderCalculater):
    def __init__(self, *args, **kwargs):
        super(ExecuteQueryTaskerCalculater, self).__init__(*args, **kwargs)

        self.config = None
        self.arguments = None
        self.executor = None
        self.tasker = None
        self.tasker_index = 0

    def start(self, tasker, loader, arguments, task_config, **kwargs):
        current_tasker = _thread_local.current_tasker
        try:
            self.config = task_config
            kn, knl = (task_config["name"] + "@"), len(task_config["name"] + "@")
            self.arguments = {}
            for key, value in arguments.items():
                if key[:knl] != kn:
                    continue
                self.arguments[key[knl:]] = value
            self.create_executor_tasker()
        finally:
            _thread_local.current_tasker = current_tasker

    def calculate(self, primary_keys, query, task_config, *args):
        current_tasker = _thread_local.current_tasker
        try:
            if self.executor is None:
                self.create_executor_tasker()
            with self.executor as executor:
                for exp, values in query["filters"].items():
                    if not exp:
                        continue
                    for key, value in values:
                        if not key or not isinstance(key, str):
                            continue
                        executor.env_variables["@" + key] = value
                database, collection_name = self.tasker.tasker.outputer.db, self.tasker.tasker.outputer.name
                executor.execute()
                query = database.query(collection_name, ["id"])
                datas = query.commit()
                delete = database.delete(collection_name, ["id"])
                delete.commit()
                return datas
        finally:
            self.executor, self.tasker = None, None
            _thread_local.current_tasker = current_tasker

    def create_executor_tasker(self):
        from ..executor import Executor
        from ..taskers.query import QueryTasker

        current_executor = Executor.current()
        if self.tasker_index <= 0:
            self.tasker_index = current_executor.distribute_tasker_index()
        with Executor(current_executor.manager, current_executor.session_config, current_executor) as executor:
            config = copy.deepcopy(self.config)
            config["output"] = "&.--.__queryTasker_" + str(id(executor)) + "::" + config["output"].split("::")[-1]
            config["name"] = config["name"] + "#queryTasker"
            tasker = QueryTasker(config, is_inner_subquery=True)
            tasker.tasker_index = self.tasker_index
            executor.runners.extend(tasker.start(config.get("name"), executor, executor.session_config,
                                                 executor.manager, self.arguments))
            self.tasker, self.executor = tasker, executor


# -*- coding: utf-8 -*-
# 2023/4/21
# create by: snower

from .example import ExampleTestCase


class DemoExampleTestCase(ExampleTestCase):
    example_name = "demo"

    def test_demo(self):
        self.execute("demo.sql")

        self.assert_data(3, [{'site_id': 8, 'site_name': '黄豆网', 'site_amount': 17.04, 'timeout_at': '16:00:00',
                              'vip_timeout_at': '11:00:00'},
                             {'site_id': 15, 'site_name': '青菜网', 'site_amount': 7.2, 'timeout_at': '15:00:00',
                              'vip_timeout_at': '11:00:00'},
                             {'site_id': 21, 'site_name': '去啥网', 'site_amount': 0, 'timeout_at': '16:00:00',
                              'vip_timeout_at': '11:00:00'},
                             {'site_id': 26, 'site_name': '汽车网', 'site_amount': 0, 'timeout_at': '16:00:00',
                              'vip_timeout_at': '11:00:00'},
                             {'site_id': 28, 'site_name': '火箭网', 'site_amount': 0, 'timeout_at': '15:00:00',
                              'vip_timeout_at': '10:00:00'},
                             {'site_id': 34, 'site_name': '卫星网', 'site_amount': 11.2, 'timeout_at': '16:40:00',
                              'vip_timeout_at': '11:20:00'}], "data error")

    def test_demo2(self):
        self.execute("demo2.sql")

        self.assert_data(2, [{'site_id': 8, 'site_name': '黄豆网'}, {'site_id': 15, 'site_name': '青菜网'},
                             {'site_id': 21, 'site_name': '去啥网'}, {'site_id': 26, 'site_name': '汽车网'},
                             {'site_id': 8, 'site_name': '黄豆网'}, {'site_id': 21, 'site_name': '去啥网'},
                             {'site_id': 26, 'site_name': '汽车网'}, {'site_id': 15, 'site_name': '青菜网'},
                             {'site_id': 28, 'site_name': '火箭网'}, {'site_id': 15, 'site_name': '青菜网'},
                             {'site_id': 28, 'site_name': '火箭网'}, {'site_id': 34, 'site_name': '卫星网'},
                             {'site_id': 34, 'site_name': '卫星网'}], "data error")

        self.assert_data(5, [{'order_id': 6, 'site_id': 34, 'amount': 11.2, 'status': 0},
                             {'order_id': 1, 'site_id': 8, 'amount': 10, 'status': 0}], "data error")

        self.assert_data(7, [{'order_id': 6, 'site_id': 34, 'amount': 1120.0},
                             {'order_id': 4, 'site_id': 28, 'amount': 470.0}], "data error")

        self.assert_data(9, [{'order_id': 6, 'site_id': 34, 'amount': 1120.0},
                             {'order_id': 1, 'site_id': 8, 'amount': 1000}], "data error")

        self.assert_data(15, [{'site_id': 8, 'site_name': '黄豆网', 'site_amount': 17.04, 'timeout_at': '16:00:00',
                              'vip_timeout_at': '11:00:00'},
                             {'site_id': 15, 'site_name': '青菜网', 'site_amount': 7.2, 'timeout_at': '15:00:00',
                              'vip_timeout_at': '11:00:00'},
                             {'site_id': 21, 'site_name': '去啥网', 'site_amount': 0, 'timeout_at': '16:00:00',
                              'vip_timeout_at': '11:00:00'},
                             {'site_id': 26, 'site_name': '汽车网', 'site_amount': 0, 'timeout_at': '16:00:00',
                              'vip_timeout_at': '11:00:00'},
                             {'site_id': 28, 'site_name': '火箭网', 'site_amount': 0, 'timeout_at': '15:00:00',
                              'vip_timeout_at': '10:00:00'},
                             {'site_id': 34, 'site_name': '卫星网', 'site_amount': 11.2, 'timeout_at': '16:40:00',
                              'vip_timeout_at': '11:20:00'}], "data error")

        self.assert_data(17, [{'site_id': 8, 'site_name': '黄豆网', 'site_amount': 17.04, 'timeout_at': '16:00:00',
                              'vip_timeout_at': '11:00:00'},
                             {'site_id': 15, 'site_name': '青菜网', 'site_amount': 7.2, 'timeout_at': '15:00:00',
                              'vip_timeout_at': '10:00:00'},
                             {'site_id': 21, 'site_name': '去啥网', 'site_amount': 0, 'timeout_at': '16:00:00',
                              'vip_timeout_at': '11:00:00'},
                             {'site_id': 26, 'site_name': '汽车网', 'site_amount': 0, 'timeout_at': '16:00:00',
                              'vip_timeout_at': '11:00:00'},
                             {'site_id': 28, 'site_name': '火箭网', 'site_amount': 0, 'timeout_at': '15:00:00',
                              'vip_timeout_at': '10:00:00'},
                             {'site_id': 34, 'site_name': '卫星网', 'site_amount': 11.2, 'timeout_at': '16:40:00',
                              'vip_timeout_at': '11:20:00'}], "data error")



# -*- coding: utf-8 -*-
# 2025/4/3
# create by: snower

import traceback
from syncany.logger import get_logger
from syncany.calculaters.calculater import Calculater


class PyEvalCalculater(Calculater):
    globals = None

    @classmethod
    def init_globals(cls):
        import sys
        import os
        import datetime
        import time
        import math
        import random
        import string
        import uuid
        import base64
        import hashlib
        import pickle
        import json
        import re
        try:
            import requests
        except:
            requests = None
        from syncany.taskers.tasker import current_tasker
        from ..executor import Executor

        cls.globals = {
            "sys": sys,
            "os": os,
            "datetime": datetime,
            "time": time,
            "math": math,
            "random": random,
            "string": string,
            "uuid": uuid,
            "base64": base64,
            "hashlib": hashlib,
            "pickle": pickle,
            "json": json,
            "re": re,
            "current_tasker": lambda : current_tasker(),
            "current_executor": lambda : Executor.current(),
            "current_manager": lambda: Executor.current().manager,
            "current_session": lambda: Executor.current().session_config,
            "current_env_variables": lambda: Executor.current().env_variables,
        }
        if requests is not None:
            cls.globals["requests"] = requests

    def __init__(self, *args, **kwargs):
        Calculater.__init__(self, *args, **kwargs)

        if self.globals is None:
            self.init_globals()
        if self.name == "pyevalt":
            self.calculate = self.calculate_eval_this

    def calculate(self, *args):
        if not args or not args[0]:
            return None
        try:
            return eval(args[0], self.globals, {"args": args[1:]})
        except Exception as e:
            get_logger().warning("pyeval calculater execute %s error: %s\n%s", args, e, traceback.format_exc())
            return None

    def calculate_eval_this(self, *args):
        if not args or len(args) < 2 or not args[0] or not args[1]:
            return None
        try:
            return eval(args[1], self.globals, {"this": args[0], "args": args[2:]})
        except Exception as e:
            get_logger().warning("pyeval calculater execute %s error: %s\n%s", args, e, traceback.format_exc())
            return None


def register_pyeval_module(name, module):
    if PyEvalCalculater.globals is None:
        PyEvalCalculater.init_globals()
    PyEvalCalculater.globals[name] = module



# -*- coding: utf-8 -*-
# 2023/2/15
# create by: snower

from syncany.hook import Hooker
from .query import QueryTasker


class StreamingFollowHooker(Hooker):
    def __init__(self, manager, tasker):
        self.manager = manager
        self.tasker = tasker

    def loaded(self, tasker, datas):
        tasker.loader.set_streaming(False)
        return datas

    def outputed(self, tasker, datas):
        tasker.loader.set_streaming(True if tasker.outputer.is_streaming() else False)


class DeleteTasker(object):
    def __init__(self, config):
        self.tasker = QueryTasker(config)

    def start(self, name, executor, session_config, manager, arguments):
        arguments["@limit"] = 0
        arguments["@batch"] = 0
        self.tasker.config["name"] = self.tasker.config["name"] + "#delete"
        taskers = self.tasker.start(name, executor, session_config, manager, arguments)
        self.tasker.tasker.add_hooker(StreamingFollowHooker(manager, self))
        return taskers

    def run(self, executor, session_config, manager):
        try:
            return self.tasker.run(executor, session_config, manager)
        finally:
            self.tasker = None

    def terminate(self):
        if not self.tasker:
            return
        self.tasker.terminate()
        self.tasker = None



# -*- coding: utf-8 -*-
# 2023/3/17
# create by: snower


class IntoTasker(object):
    def __init__(self, tasker, config):
        self.tasker = tasker
        self.config = config

    def start(self, name, executor, session_config, manager, arguments):
        for variable in self.config["variables"]:
            if variable in executor.env_variables:
                continue
            executor.env_variables[variable] = None
        self.tasker.config["output"] = "&.--.__into_" + str(id(self)) + "::" + self.tasker.config["output"].split("::")[-1]
        self.tasker.config["name"] = self.tasker.config["name"] + "#into"
        self.tasker.start(name, executor, session_config, manager, arguments)
        return [self]

    def run(self, executor, session_config, manager):
        try:
            core_tasker = self.tasker.tasker
            if not core_tasker:
                return 1
            database = core_tasker.outputer.db
            name = core_tasker.outputer.name
            schema_keys = tuple(core_tasker.schema.keys()) if isinstance(core_tasker.schema, dict) else None
            self.tasker.run(executor, session_config, manager)

            query = database.query(name, ["id"])
            datas = query.commit()
            delete = database.delete(name, ["id"])
            delete.commit()

            if not schema_keys:
                value = datas[0] if len(datas) == 1 else datas
                if len(self.config["variables"]) == 1 and isinstance(value, dict) and len(value) == 1:
                    value = list(value.values())[0]
            elif len(schema_keys) == 1:
                if len(datas) == 1:
                    value = datas[0][schema_keys[0]] if schema_keys[0] in datas[0] else None
                else:
                    value = [data[schema_keys[0]] for data in datas if schema_keys[0] in data]
            else:
                if len(datas) == 1:
                    value = {key: datas[0].get(key) for key in schema_keys}
                else:
                    value = [{key: data.get(key) for key in schema_keys} for data in datas]

            if len(self.config["variables"]) == 1:
                if value is None:
                    if self.config["variables"][0] in executor.env_variables:
                        executor.env_variables[self.config["variables"][0]] = None
                else:
                    executor.env_variables[self.config["variables"][0]] = value
            else:
                for i in range(len(self.config["variables"])):
                    if i >= len(schema_keys):
                        if self.config["variables"][i] in executor.env_variables:
                            executor.env_variables[self.config["variables"][i]] = None
                        continue
                    if isinstance(value, dict):
                        executor.env_variables[self.config["variables"][i]] = value[schema_keys[i]]
                    else:
                        executor.env_variables[self.config["variables"][i]] = [v[schema_keys[i]] for v in value]
        finally:
            self.tasker = None

    def terminate(self):
        if not self.tasker:
            return
        self.tasker.terminate()
        self.tasker = None



# -*- coding: utf-8 -*-
# 2023/4/24
# create by: snower

from .example import ExampleTestCase


class StringsExampleTestCase(ExampleTestCase):
    example_name = "strings"

    def test_strings(self):
        self.execute("strings.sql")

        self.assert_data(2, [{"a": "abc", "b": "ab", "c": "abc", "d": "ABC"}], "data error")
        self.assert_data(4, [{"a": "a b c", "b": "aaa", "c": "cba", "d": -1}], "data error")
        self.assert_data(6, [{"a": 1, "b": 1, "c": 1}], "data error")
        self.assert_data(8, [{"a": None, "b": 'a 2023-04-02 10:08:06 2023-04-02 10:08:06 1 0 1 1.23'}], "data error")



# -*- coding: utf-8 -*-
# 2023/2/7
# create by: snower

import sys
import os
import signal
import traceback
import time
from syncany.logger import get_logger
from syncany.taskers.manager import TaskerManager
from syncany.database.database import DatabaseManager
from .config import GlobalConfig
from syncanysql.executor import Executor
from .parser import SqlParser, FileParser


def main():
    if os.getcwd() not in sys.path:
        sys.path.insert(0, os.getcwd())
    try:
        fileext = os.path.splitext(sys.argv[1])[1].lower() if len(sys.argv) >= 2 else None
    except:
        fileext = None
    if sys.stdin.isatty() and fileext and fileext not in (".sql", ".sqlx", ".prql"):
        print("usage: syncany [-h] sqlx|sql|prql")
        print("syncany error: require sqlx or sql or prql file")
        exit(2)

    try:
        if sys.platform != "win32":
            signal.signal(signal.SIGHUP, lambda signum, frame: executor.terminate())
            signal.signal(signal.SIGTERM, lambda signum, frame: executor.terminate())

        global_config = GlobalConfig()
        init_execute_files = global_config.load()
        if not sys.stdin.isatty() and (not fileext or fileext not in (".sql", ".sqlx", ".prql")):
            global_config.config_logging(False)
        else:
            global_config.config_logging(True)
        global_config.load_extensions()
        manager = TaskerManager(DatabaseManager())

        try:
            with Executor(manager, global_config.session()) as executor:
                if init_execute_files:
                    for init_execute_file in init_execute_files:
                        file_parser = FileParser(init_execute_file)
                        executor.run("init " + init_execute_file, file_parser.load())
                        executor.execute()

                if not sys.stdin.isatty() and (not fileext or fileext not in (".sql", ".sqlx", ".prql")):
                    start_time = time.time()
                    content = sys.stdin.read().strip()
                    if not content:
                        return
                    sql_parser = SqlParser(content[1:-1] if content[0] in ('"', "'") and content[-1] in ('"', "'") else content)
                    sqls = sql_parser.split()
                    executor.run("pipe", sqls)
                    get_logger().info("execute pipe sql start %d task", sys.argv[1], executor.get_tasker_index())
                    try:
                        executor.execute()
                    except Exception as e:
                        get_logger().info("execute pipe sql finish with Exception %s:%s %.2fms", e.__class__.__name__, e,
                                          (time.time() - start_time) * 1000)
                        raise e
                    else:
                        get_logger().info("execute pipe sql finish %.2fms", (time.time() - start_time) * 1000)
                elif len(sys.argv) >= 2:
                    start_time = time.time()
                    file_parser = FileParser(sys.argv[1])
                    sqls = file_parser.load()
                    executor.run(sys.argv[1], sqls)
                    get_logger().info("execute file %s start %d task", sys.argv[1], executor.get_tasker_index())
                    try:
                        executor.execute()
                    except Exception as e:
                        get_logger().info("execute file %s finish with Exception %s:%s %.2fms", sys.argv[1], e.__class__.__name__, e,
                                          (time.time() - start_time) * 1000)
                        raise e
                    else:
                        get_logger().info("execute file %s finish %.2fms", sys.argv[1], (time.time() - start_time) * 1000)
                else:
                    from .prompt import CliPrompt
                    cli_prompt = CliPrompt(manager, global_config.session(), executor)
                    cli_prompt.run()
        finally:
            manager.close()
    except SystemError:
        get_logger().error("signal exited")
        exit(130)
    except KeyboardInterrupt:
        get_logger().error("Crtl+C exited")
        exit(130)
    except Exception as e:
        get_logger().error("%s\n%s", e, traceback.format_exc())
        exit(1)

if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-
# 2023/3/2
# create by: snower

import re
import time
from datetime import datetime as datetime_datetime, date as datetime_date, time as datetime_time, timedelta as datetime_timedelta
import pytz
from syncany.utils import get_timezone, parse_datetime
from syncany.calculaters import typing_filter
from ...utils import NumberTypes, NumberStringTypes

TIMEDELTA_UNITS = {"YEAR": 365 * 24 * 60 * 60, "QUARTER": 90 * 24 * 60 * 60, "MONTH": 30 * 24 * 60 * 60,
                   "WEEK": 7 * 24 * 60 * 60, "DAY": 24 * 60 * 60, "HOUR": 3600, "MINUTE": 60, "SECOND": 1,
                   "MICROSECOND": 1}
TIME_INTERVAL_RE = re.compile(r'^(\d{4}?\-)?(\d+?\-)?(\d+?\s)?(\d+?:)?(\d+?)?(:\d+)?(\.\d+)?$')

def calculate_datetime(dt, interval, is_sub=False):
    def calculate(unit, value):
        if unit == "YEAR":
            return dt.replace(year=dt.year - value) if is_sub else dt.replace(year=dt.year + value)
        if unit in ("QUARTER", "MONTH"):
            months = value * 3 if unit == "QUARTER" else value
            if is_sub:
                new_year = dt.year - int(months / 12) - (1 if dt.month - months % 12 < 1 else 0)
                new_month = (dt.month - months % 12) + (12 if dt.month - months % 12 < 1 else 0)
            else:
                new_year = dt.year + int(months / 12) + (1 if dt.month + months % 12 > 12 else 0)
                new_month = (dt.month + months % 12) % 12
            try:
                return dt.replace(year=new_year, month=new_month)
            except ValueError:
                day = dt.day - 1
                while day >= 1:
                    try:
                        return dt.replace(year=new_year, month=new_month, day=day)
                    except ValueError:
                        day -= 1
                return None
        if unit in TIMEDELTA_UNITS:
            td = datetime_timedelta(seconds=TIMEDELTA_UNITS[unit] * value)
            return (dt - td) if is_sub else (dt + td)
        return None
    
    if isinstance(interval, dict):
        dt = calculate(interval["unit"].upper(), int(interval["value"]))
        if dt:
            return dt
        interval = str(interval["value"])
    m = TIME_INTERVAL_RE.match(interval)
    if not m:
        return None
    values, units = list(m.groups()), ("YEAR", "MONTH", "DAY", "HOUR", "MINUTE", "SECOND", "MICROSECOND")
    if not values[2] and values[3] and values[4] and not values[5]:
        values[5], values[4], values[3] = values[4], values[3], None
    for i in range(7):
        if not values[i]:
            continue
        dt = calculate(units[i], int(values[i].replace("-", "").replace(".", "").replace(":", "")))
        if not dt:
            return None
    return dt


@typing_filter(datetime_datetime)
def mysql_currenttimestamp():
    return datetime_datetime.now(tz=get_timezone())

@typing_filter(datetime_date)
def mysql_curdate():
    return datetime_date.today()

@typing_filter(datetime_date)
def mysql_currentdate():
    return datetime_date.today()

@typing_filter(datetime_time)
def mysql_curtime():
    dt = datetime_datetime.now(tz=get_timezone())
    return datetime_time(dt.hour, dt.minute, dt.second, dt.microsecond)

@typing_filter(datetime_time)
def mysql_currenttime():
    dt = datetime_datetime.now(tz=get_timezone())
    return datetime_time(dt.hour, dt.minute, dt.second, dt.microsecond)

@typing_filter(datetime_datetime)
def mysql_sysdate():
    return datetime_datetime.now(tz=get_timezone())

@typing_filter(datetime_date)
def mysql_date(dt):
    if dt is None:
        return None
    if isinstance(dt, NumberStringTypes):
        dt = parse_datetime(str(dt), None, get_timezone())
    if isinstance(dt, datetime_datetime):
        return datetime_date(dt.year, dt.month, dt.day)
    if isinstance(dt, datetime_date):
        return dt
    return None

@typing_filter(datetime_datetime)
def mysql_datetime(dt):
    if dt is None:
        return None
    if isinstance(dt, NumberStringTypes):
        dt = parse_datetime(str(dt), None, get_timezone())
    if isinstance(dt, datetime_datetime):
        return dt
    if isinstance(dt, datetime_date):
        return datetime_datetime(dt.year, dt.month, dt.day, tzinfo=get_timezone())
    if isinstance(dt, datetime_time):
        now = datetime_datetime.now(tz=get_timezone())
        return datetime_datetime(now.year, now.month, now.day, dt.hour, dt.minute,
                                 dt.second, dt.microsecond, tzinfo=get_timezone())
    return None

@typing_filter(datetime_time)
def mysql_time(dt):
    if dt is None:
        return None
    if isinstance(dt, NumberStringTypes):
        dt = parse_datetime(str(dt), None, get_timezone())
    if isinstance(dt, datetime_time):
        return dt
    if isinstance(dt, datetime_datetime):
        return datetime_time(dt.hour, dt.minute, dt.second, dt.microsecond, tzinfo=get_timezone())
    if isinstance(dt, datetime_date):
        return datetime_time(tzinfo=get_timezone())
    return None

@typing_filter(int)
def mysql_unix_timestamp(dt=None):
    if dt is None:
        return int(time.time())
    return int(time.mktime(dt.utctimetuple()))

@typing_filter(datetime_datetime)
def mysql_from_unixtime(t):
    if t is None:
        return None
    return datetime_datetime.fromtimestamp(int(t), pytz.UTC).astimezone(tz=get_timezone())

@typing_filter(int)
def mysql_month(dt):
    if dt is None:
        return None
    if isinstance(dt, NumberStringTypes):
        dt = parse_datetime(str(dt), None, get_timezone())
    return dt.month

@typing_filter(str)
def mysql_monthname(dt):
    if dt is None:
        return None
    if isinstance(dt, NumberStringTypes):
        dt = parse_datetime(str(dt), None, get_timezone())
    return dt.strftime("%b")

@typing_filter(str)
def mysql_dayname(dt):
    if dt is None:
        return None
    if isinstance(dt, NumberStringTypes):
        dt = parse_datetime(str(dt), None, get_timezone())
    return dt.strftime("%A")

@typing_filter(int)
def mysql_dayofweek(dt):
    if dt is None:
        return None
    if isinstance(dt, NumberStringTypes):
        dt = parse_datetime(str(dt), None, get_timezone())
    return dt.weekday() + 1

@typing_filter(int)
def mysql_week(dt, mod=None):
    if dt is None:
        return None
    if isinstance(dt, NumberStringTypes):
        dt = parse_datetime(str(dt), None, get_timezone())
    return int(dt.strftime("%W" if str(mod) == "1" else "%U"))

@typing_filter(int)
def mysql_yearweek(dt, mod=None):
    if dt is None:
        return None
    if isinstance(dt, NumberStringTypes):
        dt = parse_datetime(str(dt), None, get_timezone())
    return int(dt.strftime("%Y%W" if str(mod) == "1" else "%Y%U"))

@typing_filter(int)
def mysql_dayofyear(dt):
    if dt is None:
        return None
    if isinstance(dt, NumberStringTypes):
        dt = parse_datetime(str(dt), None, get_timezone())
    return int(dt.strftime("%j"))

@typing_filter(int)
def mysql_dayofmonth(dt):
    if dt is None:
        return None
    if isinstance(dt, NumberStringTypes):
        dt = parse_datetime(str(dt), None, get_timezone())
    return dt.day

@typing_filter(int)
def mysql_year(dt):
    if dt is None:
        return None
    if isinstance(dt, NumberStringTypes):
        dt = parse_datetime(str(dt), None, get_timezone())
    return dt.year

@typing_filter(int)
def mysql_time_to_sec(dt):
    if dt is None:
        return None
    st = parse_datetime("2000-01-01 00:00:00", None, get_timezone())
    if isinstance(dt, str):
        dt = parse_datetime("2000-01-01 " + dt, None, get_timezone())
    return int((dt - st).total_seconds())

@typing_filter(str)
def mysql_sec_to_time(t):
    if t is None:
        return None
    st = parse_datetime("2000-01-01 00:00:00", None, get_timezone())
    return (st + datetime_timedelta(seconds=t)).strftime("%H:%M:%S")

@typing_filter(datetime_datetime)
def mysql_dateadd(dt, i):
    if dt is None:
        return None
    if isinstance(dt, NumberStringTypes):
        dt = parse_datetime(str(dt), None, get_timezone())
    if isinstance(i, NumberTypes) or (isinstance(i, str) and i.isdigit()):
        return dt + datetime_timedelta(days=int(i))
    return calculate_datetime(dt, i, False)

@typing_filter(datetime_datetime)
def mysql_adddate(dt, i):
    if dt is None:
        return None
    if isinstance(dt, NumberStringTypes):
        dt = parse_datetime(str(dt), None, get_timezone())
    if isinstance(i, NumberTypes) or (isinstance(i, str) and i.isdigit()):
        return dt + datetime_timedelta(days=int(i))
    return calculate_datetime(dt, i, False)

@typing_filter(datetime_datetime)
def mysql_datesub(dt, i):
    if dt is None:
        return None
    if isinstance(dt, NumberStringTypes):
        dt = parse_datetime(str(dt), None, get_timezone())
    if isinstance(i, NumberTypes) or (isinstance(i, str) and i.isdigit()):
        return dt - datetime_timedelta(days=int(i))
    return calculate_datetime(dt, i, True)

@typing_filter(datetime_datetime)
def mysql_subdate(dt, i):
    if dt is None:
        return None
    if isinstance(dt, NumberStringTypes):
        dt = parse_datetime(str(dt), None, get_timezone())
    if isinstance(i, NumberTypes) or (isinstance(i, str) and i.isdigit()):
        return dt - datetime_timedelta(days=int(i))
    return calculate_datetime(dt, i, True)

@typing_filter(datetime_datetime)
def mysql_addtime(dt, i):
    if dt is None:
        return None
    if isinstance(dt, NumberStringTypes):
        dt = parse_datetime(str(dt), None, get_timezone())
    if isinstance(i, NumberTypes) or (isinstance(i, str) and i.isdigit()):
        return dt + datetime_timedelta(days=int(i))
    return calculate_datetime(dt, i, False)

@typing_filter(datetime_datetime)
def mysql_subtime(dt, i):
    if dt is None:
        return None
    if isinstance(dt, NumberStringTypes):
        dt = parse_datetime(str(dt), None, get_timezone())
    if isinstance(i, NumberTypes) or (isinstance(i, str) and i.isdigit()):
        return dt - datetime_timedelta(days=int(i))
    return calculate_datetime(dt, i, True)

@typing_filter(int)
def mysql_datediff(dt1, dt2):
    if dt1 is None or dt2 is None:
        return None
    if isinstance(dt1, NumberStringTypes):
        dt1 = parse_datetime(str(dt1), None, get_timezone())
    if isinstance(dt2, NumberStringTypes):
        dt2 = parse_datetime(str(dt2), None, get_timezone())
    return int((dt2 - dt1).total_seconds() / 86400)

@typing_filter(int)
def mysql_timestampdiff(unit, dt1, dt2):
    if dt1 is None or dt2 is None:
        return None
    if isinstance(dt1, NumberStringTypes):
        dt1 = parse_datetime(str(dt1), None, get_timezone())
    if isinstance(dt2, NumberStringTypes):
        dt2 = parse_datetime(str(dt2), None, get_timezone())
    return int((dt2 - dt1).total_seconds() / TIMEDELTA_UNITS[unit])

@typing_filter(str)
def mysql_date_format(dt, f):
    if dt is None:
        return None
    if isinstance(dt, NumberStringTypes):
        dt = parse_datetime(str(dt), None, get_timezone())
    return dt.strftime(f.replace("%v", "%V"))

@typing_filter(str)
def mysql_time_format(dt, f):
    if dt is None:
        return None
    if isinstance(dt, NumberStringTypes):
        dt = parse_datetime("2000-01-01 " + str(dt), None, get_timezone())
    return dt.strftime(f.replace("%v", "%V"))

@typing_filter(int)
def mysql_weekday(dt):
    if dt is None:
        return None
    if isinstance(dt, NumberStringTypes):
        dt = parse_datetime(str(dt), None, get_timezone())
    return dt.weekday()

@typing_filter(datetime_date)
def mysql_utc_date():
    dt = datetime_datetime.now(pytz.UTC)
    return datetime_date(dt.year, dt.month, dt.day)

@typing_filter(datetime_time)
def mysql_utc_time():
    dt = datetime_datetime.now(pytz.UTC)
    return datetime_time(dt.hour, dt.minute, dt.second, dt.microsecond)

@typing_filter(datetime_datetime)
def mysql_utc_timestamp():
    return datetime_datetime.now(pytz.UTC)


funcs = {key[6:]: value for key, value in globals().items() if key.startswith("mysql_")}



# -*- coding: utf-8 -*-
# 2023/3/21
# create by: snower

import datetime
import json
from syncany.calculaters.calculater import Calculater
from syncany.filters import IntFilter, FloatFilter, ArrayFilter, StringFilter
from ..utils import NumberDecimalTypes, ensure_number, ensure_int


class AggregateKeyCalculater(Calculater):
    def calculate(self, *args):
        if len(args) == 1:
            return tuple(args[0]) if isinstance(args[0], list) else args[0]
        return args


class AggregateCalculater(Calculater):
    def __init__(self, *args, **kwargs):
        super(AggregateCalculater, self).__init__(*args, **kwargs)
        
        if self.name.endswith("::aggregate"):
            self.calculate = self.aggregate
        elif self.name.endswith("::reduce"):
            self.calculate = self.reduce
        
    def aggregate(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            return data_value
        return state_value + data_value
    
    def reduce(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            return data_value
        return state_value + data_value
    
    def calculate(self, *args):
        if self.name.endswith("::aggregate"):
            return self.aggregate(*args)
        if self.name.endswith("::reduce"):
            return self.reduce(*args)
        return None
    
    
class StateAggregateCalculater(AggregateCalculater):
    def __init__(self, *args, **kwargs):
        super(StateAggregateCalculater, self).__init__(*args, **kwargs)

        if self.name.endswith("::final_value"):
            self.calculate = self.final_value
            
    def final_value(self, state_value):
        return state_value
    
    def calculate(self, *args):
        if self.name.endswith("::final_value"):
            return self.final_value(*args)
        return super(StateAggregateCalculater, self).calculate(*args)


class AggregateCountCalculater(AggregateCalculater):
    def aggregate(self, state_value, data_value):
        if data_value is None:
            return state_value or 0
        try:
            return state_value + 1
        except:
            if state_value is None:
                return 1
            try:
                return int(state_value) + 1
            except:
                return state_value

    def get_final_filter(self):
        return IntFilter.default()

class AggregateDistinctCountCalculater(StateAggregateCalculater):
    def aggregate(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            try:
                return {data_value}
            except:
                if isinstance(data_value, list):
                    try:
                        return {tuple(data_value)}
                    except:
                        return {str(data_value)}
                return {str(data_value)}
        try:
            state_value.add(data_value)
        except:
            if isinstance(data_value, list):
                try:
                    state_value.add(tuple(data_value))
                except:
                    state_value.add(str(data_value))
            else:
                state_value.add(str(data_value))
        return state_value

    def reduce(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            return data_value
        return state_value | data_value

    def final_value(self, state_value):
        if state_value is None:
            return 0
        return len(state_value)

    def get_final_filter(self):
        return IntFilter.default()


class AggregateSumCalculater(AggregateCalculater):
    def aggregate(self, state_value, data_value):
        try:
            if isinstance(data_value, NumberDecimalTypes):
                return state_value + data_value
            return state_value + ensure_number(data_value)
        except:
            if data_value is None:
                return state_value or 0
            if state_value is None:
                if isinstance(data_value, NumberDecimalTypes):
                    return data_value
                return ensure_number(data_value)
            return state_value

    def get_final_filter(self):
        return FloatFilter.default()


class AggregateMaxCalculater(AggregateCalculater):
    def aggregate(self, state_value, data_value):
        try:
            return max(state_value, data_value)
        except:
            if data_value is None:
                return state_value
            if state_value is None:
                return data_value
            try:
                return max(state_value, type(state_value)(data_value))
            except:
                return state_value

    def reduce(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            return data_value
        return max(state_value, data_value)


class AggregateMinCalculater(AggregateCalculater):
    def aggregate(self, state_value, data_value):
        try:
            return min(state_value, data_value)
        except:
            if data_value is None:
                return state_value
            if state_value is None:
                return data_value
            try:
                return min(state_value, type(state_value)(data_value))
            except:
                return state_value

    def reduce(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            return data_value
        return min(state_value, data_value)


class AggregateAvgCalculater(StateAggregateCalculater):
    def aggregate(self, state_value, data_value):
        try:
            state_value["count_value"] += 1
            if isinstance(data_value, NumberDecimalTypes):
                state_value["sum_value"] += data_value
            else:
                state_value["sum_value"] += ensure_number(data_value)
            return state_value
        except:
            if data_value is None:
                return state_value
            if state_value is None:
                if data_value is None:
                    return None
                if isinstance(data_value, NumberDecimalTypes):
                    return {"count_value": 1, "sum_value": data_value}
                return {"count_value": 1, "sum_value": ensure_number(data_value)}
            return state_value

    def reduce(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            return data_value
        return {"count_value": state_value["count_value"] + data_value["count_value"],
                "sum_value": state_value["sum_value"] + data_value["sum_value"]}

    def final_value(self, state_value):
        if state_value is None:
            return 0
        return state_value["sum_value"] / state_value["count_value"]

    def get_final_filter(self):
        return FloatFilter.default()


class AggregateGroupConcatCalculater(StateAggregateCalculater):
    def aggregate(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            return [str(data_value)]
        state_value.append(str(data_value))
        return state_value

    def reduce(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            return data_value
        return state_value + data_value

    def final_value(self, state_value):
        if not state_value:
            return ""
        return ",".join(state_value)


class AggregateGroupArrayCalculater(StateAggregateCalculater):
    def aggregate(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            return [data_value]
        state_value.append(data_value)
        return state_value

    def reduce(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            return data_value
        return state_value + data_value

    def final_value(self, state_value):
        if state_value is None:
            return []
        return state_value

    def get_final_filter(self):
        return FloatFilter.default()


class AggregateGroupUniqArrayCalculater(StateAggregateCalculater):
    def aggregate(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            try:
                return {data_value}
            except Exception as e:
                if isinstance(data_value, list):
                    return {tuple(data_value)}
                raise e
        try:
            state_value.add(data_value)
        except Exception as e:
            if isinstance(data_value, list):
                state_value.add(tuple(data_value))
            else:
                raise e
        return state_value

    def reduce(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            return data_value
        return state_value | data_value

    def final_value(self, state_value):
        if state_value is None:
            return []
        return list(state_value)

    def get_final_filter(self):
        return ArrayFilter.default()


class AggregateGroupBitAndCalculater(AggregateCalculater):
    def aggregate(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            return data_value
        if isinstance(data_value, int):
            return state_value & data_value
        return state_value & ensure_int(data_value)

    def reduce(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            return data_value
        return state_value & data_value

    def get_final_filter(self):
        return IntFilter.default()


class AggregateGroupBitOrCalculater(AggregateCalculater):
    def aggregate(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            return data_value
        if isinstance(data_value, int):
            return state_value | data_value
        return state_value | ensure_int(data_value)

    def reduce(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            return data_value
        return state_value | data_value

    def get_final_filter(self):
        return IntFilter.default()


class AggregateGroupBitXorCalculater(AggregateCalculater):
    def aggregate(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            return data_value
        if isinstance(data_value, int):
            return state_value ^ data_value
        return state_value ^ ensure_int(data_value)

    def reduce(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            return data_value
        return state_value ^ data_value

    def get_final_filter(self):
        return IntFilter.default()


class AggregateJsonArrayaggCalculater(StateAggregateCalculater):
    def format_value(self, value):
        if isinstance(value, datetime.date):
            if isinstance(value, datetime.datetime):
                return value.strftime("%Y-%m-%d %H:%M:%S")
            return value.strftime("%Y-%m-%d")
        if isinstance(value, datetime.time):
            return value.strftime("%H:%M:%S")
        return str(value)

    def aggregate(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            return [data_value]
        state_value.append(data_value)
        return state_value

    def reduce(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            return data_value
        return state_value + data_value

    def final_value(self, state_value):
        if state_value is None:
            return None
        return json.dumps(state_value, ensure_ascii=False, default=self.format_value)

    def get_final_filter(self):
        return StringFilter.default()


class AggregateJsonObjectaggCalculater(StateAggregateCalculater):
    def format_value(self, value):
        if isinstance(value, datetime.date):
            if isinstance(value, datetime.datetime):
                return value.strftime("%Y-%m-%d %H:%M:%S")
            return value.strftime("%Y-%m-%d")
        if isinstance(value, datetime.time):
            return value.strftime("%H:%M:%S")
        return str(value)

    def aggregate(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            state_value = {}
        if isinstance(data_value, list):
            if len(data_value) >= 2:
                state_value[self.format_value(data_value[0])] = data_value[1]
            elif data_value:
                state_value[self.format_value(data_value[0])] = None
        else:
            state_value[self.format_value(data_value)] = None
        return state_value

    def reduce(self, state_value, data_value):
        if data_value is None:
            return state_value
        if state_value is None:
            return data_value
        return state_value.update(data_value)

    def final_value(self, state_value):
        if state_value is None:
            return None
        return json.dumps(state_value, ensure_ascii=False, default=self.format_value)

    def get_final_filter(self):
        return StringFilter.default()


class AggregateAnyValueCalculater(StateAggregateCalculater):
    def aggregate(self, state_value, data_value):
        if state_value is None:
            return {"value": data_value}
        return state_value

    def reduce(self, state_value, data_value):
        if state_value is None:
            return data_value
        return state_value

    def final_value(self, state_value):
        if state_value is None:
            return None
        return state_value["value"]


# -*- coding: utf-8 -*-
# 2023/4/21
# create by: snower

import datetime
from .example import ExampleTestCase


class DatetimeExampleTestCase(ExampleTestCase):
    example_name = "datetime"

    def test_datetime(self):
        self.execute("datetime.sql")

        self.assert_value(2, "NOW()", lambda value: isinstance(value, datetime.datetime), "data error")

        self.assert_value(4, "NOW(0)", lambda value: isinstance(value, datetime.datetime), "data error")
        self.assert_value(4, "NOW('-1d')", lambda value: isinstance(value, datetime.datetime), "data error")
        self.assert_value(4, "NOW('+3d')", lambda value: isinstance(value, datetime.datetime), "data error")
        self.assert_value(4, "NOW('-3d', 0)", lambda value: isinstance(value, datetime.datetime), "data error")
        self.assert_value(4, "NOW('-3d', 0, 10, 11)", lambda value: isinstance(value, datetime.datetime), "data error")

        self.assert_value(6, "DATE_ADD(NOW(), 1, '')", lambda value: isinstance(value, datetime.datetime), "data error")
        self.assert_value(6, "ADDDATE(NOW(), INTERVAL '20' DAY)", lambda value: isinstance(value, datetime.datetime),
                          "data error")
        self.assert_value(6, "DATE_SUB(NOW(), INTERVAL '10' DAY)", lambda value: isinstance(value, datetime.datetime),
                          "data error")
        self.assert_value(6, "SUBDATE(NOW(), INTERVAL '7' MONTH)", lambda value: isinstance(value, datetime.datetime),
                          "data error")

        self.assert_value(8, "ADDTIME(NOW(), '10:00')", lambda value: isinstance(value, datetime.datetime),
                          "data error")
        self.assert_value(8, "SUBTIME(NOW(), '1 10:00')", lambda value: isinstance(value, datetime.datetime),
                          "data error")

        self.assert_value(10, "DATE_FORMAT(DATETIME('2023-04-24 17:07:08'), '%Y-%m-%d %H:%M:%S')",
                          '2023-04-24 17:07:08', "data error")
        self.assert_value(10, "TIME_FORMAT(DATETIME('2023-04-24 17:07:08'), '%H:%M:%S')", '17:07:08', "data error")
        self.assert_value(10, "TIME_TO_SEC('10:11:00')", 36660, "data error")
        self.assert_value(10, "SEC_TO_TIME(234)", '00:03:54', "data error")

        self.assert_value(12, "CURDATE()", lambda value: isinstance(value, datetime.date), "data error")
        self.assert_value(12, "CURRENT_DATE", lambda value: isinstance(value, datetime.date), "data error")
        self.assert_value(12, "CURRENT_TIME()", lambda value: isinstance(value, datetime.time), "data error")
        self.assert_value(12, "CURTIME()", lambda value: isinstance(value, datetime.time), "data error")

        self.assert_value(14, "FROM_UNIXTIME(1677833819)", lambda value: isinstance(value, datetime.datetime),
                          "data error")
        self.assert_value(14, "UNIX_TIMESTAMP()", lambda value: isinstance(value, int), "data error")
        self.assert_value(14, "UNIX_TIMESTAMP(NOW())", lambda value: isinstance(value, int), "data error")
        self.assert_value(14, "CURRENT_TIMESTAMP()", lambda value: isinstance(value, datetime.datetime), "data error")

        self.assert_value(16, "UTC_DATE()", lambda value: isinstance(value, datetime.date), "data error")
        self.assert_value(16, "UTC_TIME()", lambda value: isinstance(value, datetime.time), "data error")
        self.assert_value(16, "UTC_TIMESTAMP()", lambda value: isinstance(value, datetime.datetime), "data error")

        self.assert_value(18, "DATE(NOW())", lambda value: isinstance(value, datetime.date), "data error")
        self.assert_value(18, "DATETIME(NOW())", lambda value: isinstance(value, datetime.datetime), "data error")
        self.assert_value(18, "TIME(NOW())", lambda value: isinstance(value, datetime.time), "data error")
        self.assert_value(18, "DATETIME(DATE(NOW()))", lambda value: isinstance(value, datetime.datetime), "data error")
        self.assert_value(18, "DATETIME(TIME(NOW()))", lambda value: isinstance(value, datetime.datetime), "data error")
        self.assert_value(18, "DATE(TIME(NOW()))", None, "data error")
        self.assert_value(18, "TIME(DATE(NOW()))", lambda value: isinstance(value, datetime.time), "data error")

        self.assert_value(20, "DATE_FORMAT(CONVERT_DATETIME('1600-01-01'), '%Y-%m-%d %H:%M:%S')",
                          '1600-01-01 00:00:00', "data error")
        self.assert_value(20, "DATE_FORMAT(CONVERT_DATETIME('1900-01-01 10:12:23'), '%Y-%m-%d %H:%M:%S')",
                          '1900-01-01 10:12:23', "data error")
        self.assert_value(20, "TIME_FORMAT(CONVERT_DATETIME('1800-01-01 10:12:23'), '%H:%M:%S')", '10:12:23', "data error")



# -*- coding: utf-8 -*-
# 2023/5/31
# create by: snower

from .example import ExampleTestCase


class JoinsExampleTestCase(ExampleTestCase):
    example_name = "joins"

    def test_left_join(self):
        self.execute("left_join.sql")

        self.assert_data(2, [{'order_id': 1, 'name': '李四', 'goods_name': '青菜'},
                             {'order_id': 2, 'name': '王五', 'goods_name': '青菜'},
                             {'order_id': 3, 'name': '李四', 'goods_name': '白菜'},
                             {'order_id': 4, 'name': '王五', 'goods_name': '青菜'},
                             {'order_id': 5, 'name': '李四', 'goods_name': '青菜'},
                             {'order_id': 6, 'name': '李四', 'goods_name': '白菜'}], "data error")

        self.assert_data(7, [{'order_id': 1, 'name': '李四', 'goods_name': '青菜'},
                             {'order_id': 2, 'name': '王五', 'goods_name': '青菜'},
                             {'order_id': 3, 'name': '李四', 'goods_name': '白菜'},
                             {'order_id': 4, 'name': '王五', 'goods_name': '青菜'},
                             {'order_id': 5, 'name': '李四', 'goods_name': '青菜'},
                             {'order_id': 6, 'name': '李四', 'goods_name': '白菜'}], "data error")

        self.assert_data(12, [{'order_id': 1, 'name': '李四', 'goods_name': '青菜', 'cnt': 4},
                              {'order_id': 2, 'name': '王五', 'goods_name': '青菜', 'cnt': 2}], "data error")

        self.assert_data(18, [{'order_id': 1, 'name': '李四', 'goods_name': '青菜', 'ucnt': 2, 'SUM(a.amount)': 33.2},
                              {'order_id': 3, 'name': '李四', 'goods_name': '白菜', 'ucnt': 1, 'SUM(a.amount)': 10}],
                         "data error")

        self.assert_data(24, [{'name': '李四', 'goods_name': '青菜'}, {'name': '王五', 'goods_name': '青菜'},
                              {'name': '李四', 'goods_name': '白菜'}], "data error")

        self.assert_data(29, [{'code': '123', 'name': '王五', 'goods_name': '青菜'},
                              {'code': '124', 'name': '李四', 'goods_name': '青菜'},
                              {'code': '125', 'name': '李四', 'goods_name': '青菜'}], "data error")

        self.assert_data(35, [{'order_id': 1, 'name': 'ServiceA'}, {'order_id': 1, 'name': 'ServiceA'},
                              {'order_id': 2, 'name': 'ServiceA'}, {'order_id': 2, 'name': 'ServiceB'}], "data error")

        self.assert_data(40, [{'code': '123', 'name': '王五', 'goods_name': '青菜'},
                              {'code': '124', 'name': '李四', 'goods_name': '青菜'},
                              {'code': '125', 'name': '李四', 'goods_name': '青菜'}], "data error")

        self.assert_data(46, [{'cnt': 2, 'goods_name': '青菜', 'history_type': 1, 'name': '李四', 'order_id': 1,
                               'total_amount': 28.700000000000003},
                              {'cnt': 2, 'goods_name': '白菜', 'history_type': 1, 'name': '李四', 'order_id': 3,
                               'total_amount': 7.3},
                              {'cnt': 3, 'goods_name': '白菜', 'history_type': 0, 'name': '李四', 'order_id': 3,
                               'total_amount': 37.4},
                              {'cnt': 2, 'goods_name': '青菜', 'history_type': 1, 'name': '李四', 'order_id': 5,
                               'total_amount': 89.9}], "data error")

        self.assert_data(52, [{'cnt': 2, 'goods_name': '青菜', 'history_type': 1, 'name': '李四', 'order_id': 1,
                               'total_amount': 28.700000000000003},
                              {'cnt': 2, 'goods_name': '白菜', 'history_type': 1, 'name': '李四', 'order_id': 3,
                               'total_amount': 7.3},
                              {'cnt': 3, 'goods_name': '白菜', 'history_type': 0, 'name': '李四', 'order_id': 3,
                               'total_amount': 37.4},
                              {'cnt': 2, 'goods_name': '青菜', 'history_type': 1, 'name': '李四', 'order_id': 5,
                               'total_amount': 89.9}], "data error")

    def test_right_join(self):
        self.execute("right_join.sql")

        self.assert_data(2, [{'order_id': 1, 'name': '李四', 'goods_name': '青菜'},
                             {'order_id': 2, 'name': '王五', 'goods_name': '青菜'},
                             {'order_id': 3, 'name': '李四', 'goods_name': '白菜'},
                             {'order_id': 4, 'name': '王五', 'goods_name': '青菜'},
                             {'order_id': 5, 'name': '李四', 'goods_name': '青菜'},
                             {'order_id': 6, 'name': '李四', 'goods_name': '白菜'}], "data error")

        self.assert_data(7, [{'order_id': 1, 'name': '李四', 'goods_name': '青菜', 'cnt': 4},
                             {'order_id': 2, 'name': '王五', 'goods_name': '青菜', 'cnt': 2}], "data error")

        self.assert_data(12, [{'order_id': 1, 'name': '李四', 'goods_name': '青菜', 'cnt': 2, 'SUM(b.amount)': 33.2},
                              {'order_id': 3, 'name': '李四', 'goods_name': '白菜', 'cnt': 1, 'SUM(b.amount)': 10}],
                         "data error")

    def test_inner_join(self):
        self.execute("inner_join.sql")

        self.assert_data(2, [{'goods_name': '青菜', 'order_id': 1, 'name': '李四'},
                             {'goods_name': '青菜', 'order_id': 2, 'name': '王五'},
                             {'goods_name': '青菜', 'order_id': 4, 'name': '王五'},
                             {'goods_name': '青菜', 'order_id': 5, 'name': '李四'},
                             {'goods_name': '白菜', 'order_id': 3, 'name': '李四'},
                             {'goods_name': '白菜', 'order_id': 6, 'name': '李四'},
                             {'goods_name': '萝卜', 'order_id': None, 'name': None}], "data error")

        self.assert_data(7, [{'goods_name': '青菜', 'order_id': 1, 'name': '李四'},
                             {'goods_name': '青菜', 'order_id': 2, 'name': '王五'},
                             {'goods_name': '青菜', 'order_id': 4, 'name': '王五'},
                             {'goods_name': '青菜', 'order_id': 5, 'name': '李四'},
                             {'goods_name': '白菜', 'order_id': 3, 'name': '李四'},
                             {'goods_name': '白菜', 'order_id': 6, 'name': '李四'}], "data error")

        self.assert_data(12, [{'goods_name': '青菜', 'latest_order_id': 5, 'name': '李四',
                               'SUM(CASE WHEN NOT b.order_id IS NULL THEN 1 ELSE 0 END)': 4, 'user_cnt': 2,
                               'total_amount': 33.2}, {'goods_name': '白菜', 'latest_order_id': 6, 'name': '李四',
                                                       'SUM(CASE WHEN NOT b.order_id IS NULL THEN 1 ELSE 0 END)': 2,
                                                       'user_cnt': 1, 'total_amount': 10},
                              {'goods_name': '萝卜', 'latest_order_id': None, 'name': None,
                               'SUM(CASE WHEN NOT b.order_id IS NULL THEN 1 ELSE 0 END)': 0, 'user_cnt': 0,
                               'total_amount': 0}], "data error")

        self.assert_data(18, [{'goods_name': '青菜', 'latest_order_id': 6, 'name': '李四',
                               'SUM(CASE WHEN NOT b.order_id IS NULL THEN 1 ELSE 0 END)': 4},
                              {'goods_name': '青菜', 'latest_order_id': 4, 'name': '王五',
                               'SUM(CASE WHEN NOT b.order_id IS NULL THEN 1 ELSE 0 END)': 2}], "data error")

        self.assert_data(24, [{'goods_name': '青菜', 'order_id': 1, 'name': '李四'},
                             {'goods_name': '青菜', 'order_id': 2, 'name': '王五'},
                             {'goods_name': '青菜', 'order_id': 4, 'name': '王五'},
                             {'goods_name': '青菜', 'order_id': 5, 'name': '李四'},
                             {'goods_name': '白菜', 'order_id': 3, 'name': '李四'},
                             {'goods_name': '白菜', 'order_id': 6, 'name': '李四'}], "data error")

        self.assert_data(28, [{'goods_name': '青菜', 'order_id': 1, 'name': '李四'},
                              {'goods_name': '青菜', 'order_id': 2, 'name': '王五'},
                              {'goods_name': '青菜', 'order_id': 4, 'name': '王五'},
                              {'goods_name': '青菜', 'order_id': 5, 'name': '李四'},
                              {'goods_name': '白菜', 'order_id': 3, 'name': '李四'},
                              {'goods_name': '白菜', 'order_id': 6, 'name': '李四'}], "data error")



# -*- coding: utf-8 -*-
# 2023/4/21
# create by: snower

import uuid
from bson.objectid import ObjectId
from .example import ExampleTestCase


class FunctionsExampleTestCase(ExampleTestCase):
    example_name = "functions"

    def test_generate(self):
        self.execute("generate.sql")

        self.assert_value(2, "a", lambda value: isinstance(value, ObjectId), "data error")
        self.assert_value(2, "b", ObjectId('65bb4211eda4fed2e199073e'), "data error")
        self.assert_value(2, "c", lambda value: isinstance(value, uuid.UUID), "data error")
        self.assert_value(2, "d", uuid.UUID('54aa0a5c-b54f-4628-8391-3756007d5fc3'), "data error")
        self.assert_value(2, "e", lambda value: isinstance(value, int), "data error")
        self.assert_value(2, "f", lambda value: isinstance(value, int), "data error")

        self.assert_value(4, "a", lambda value: isinstance(value, float), "data error")
        self.assert_value(4, "b", lambda value: isinstance(value, int), "data error")
        self.assert_value(4, "c", lambda value: isinstance(value, str), "data error")
        self.assert_value(4, "d", lambda value: isinstance(value, str), "data error")
        self.assert_value(4, "e", lambda value: isinstance(value, str), "data error")
        self.assert_value(4, "f", lambda value: isinstance(value, str), "data error")
        self.assert_value(4, "g", lambda value: isinstance(value, str), "data error")
        self.assert_value(4, "h", lambda value: isinstance(value, bytes), "data error")

        self.assert_value(6, "a", lambda value: isinstance(value, int), "data error")



# -*- coding: utf-8 -*-
# 2023/4/24
# create by: snower

from .example import ExampleTestCase


class MathematicalExampleTestCase(ExampleTestCase):
    example_name = "mathematical"

    def test_mathematical(self):
        self.execute("mathematical.sql")

        self.assert_data(2, [{"a": 3, "b": 1, "c": 2, "d": 2}], "data error")
        self.assert_data(4, [{"a": 0, "b": 3, "c": 3, "d": -3}], "data error")
        self.assert_data(6, [{"a": 3, "b": 2022, "c": 222}], "data error")
        self.assert_data(8, [{"a": None, "b": 2, "c": 1}], "data error")
        self.assert_data(10, [{"a": 10115201050403, "b": 10115201, "c": 50403}], "data error")
        self.assert_data(12, [{"a": [2, 3, 4], "b": [4, 8, 12]}], "data error")
        self.assert_data(14, [{"a": 5, "b": 3}], "data error")



# -*- coding: utf-8 -*-
# 2023/4/21
# create by: snower

from .example import ExampleTestCase


class NginxLogExampleTestCase(ExampleTestCase):
    example_name = "nginx-log"

    def test_ip_top3(self):
        self.execute("ip-top-3.sql")

        self.assert_data(3, [{'cnt': 22, 'ip': '54.37.79.75'}, {'cnt': 14, 'ip': '143.110.222.166'},
                             {'cnt': 9, 'ip': '35.216.169.119'}], "data error")



# -*- coding: utf-8 -*-
# 2023/2/8
# create by: snower

import os
import sys
from pygments.lexers import SqlLexer
from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.lexers import PygmentsLexer
from prompt_toolkit.styles import Style
from prompt_toolkit.filters import Condition
from prompt_toolkit.cursor_shapes import CursorShape
from .version import version
from .parser import SqlParser, SqlSegment
from .calculaters import mysql_funcs

SQL_COMPLETER_WORDS = [
    "abort",
    "action",
    "add",
    "after",
    "all",
    "alter",
    "analyze",
    "and",
    "as",
    "asc",
    "attach",
    "autoincrement",
    "before",
    "begin",
    "between",
    "by",
    "cascade",
    "case",
    "cast",
    "check",
    "collate",
    "column",
    "commit",
    "conflict",
    "constraint",
    "create",
    "cross",
    "current_date",
    "current_time",
    "current_timestamp",
    "database",
    "default",
    "deferrable",
    "deferred",
    "delete",
    "desc",
    "detach",
    "distinct",
    "drop",
    "each",
    "else",
    "end",
    "escape",
    "except",
    "exclusive",
    "exists",
    "explain",
    "fail",
    "for",
    "foreign",
    "from",
    "full",
    "glob",
    "group",
    "having",
    "if",
    "ignore",
    "immediate",
    "in",
    "index",
    "indexed",
    "initially",
    "inner",
    "insert",
    "instead",
    "intersect",
    "into",
    "is",
    "isnull",
    "join",
    "key",
    "left",
    "like",
    "limit",
    "match",
    "natural",
    "no",
    "not",
    "notnull",
    "null",
    "of",
    "offset",
    "on",
    "or",
    "order",
    "outer",
    "plan",
    "pragma",
    "primary",
    "query",
    "raise",
    "recursive",
    "references",
    "regexp",
    "reindex",
    "release",
    "rename",
    "replace",
    "restrict",
    "right",
    "rollback",
    "row",
    "savepoint",
    "select",
    "set",
    "table",
    "temp",
    "temporary",
    "then",
    "to",
    "transaction",
    "trigger",
    "union",
    "unique",
    "update",
    "using",
    "vacuum",
    "values",
    "view",
    "virtual",
    "when",
    "where",
    "with",
    "without",
]

style = Style.from_dict(
    {
        "completion-menu.completion": "bg:#008888 #ffffff",
        "completion-menu.completion.current": "bg:#00aaaa #000000",
        "scrollbar.background": "bg:#88aaaa",
        "scrollbar.button": "bg:#222222",
    }
)


class CliFileHistory(FileHistory):
    def load_history_strings(self):
        strings, time_strings, lines, time_line = [], [], [], ''
        if os.path.exists(self.filename):
            with open(self.filename, "rb") as f:
                for line_bytes in f:
                    line = line_bytes.decode("utf-8", errors="replace")
                    if line.startswith("#"):
                        time_line = line
                    elif line.startswith("+"):
                        lines.append(line[1:])
                        continue
                    if not lines:
                        continue
                    string, lines = "".join(lines)[:-1], []
                    strings.append(string)
                    time_strings.append(time_line)

                if lines:
                    string, lines = "".join(lines)[:-1], []
                    strings.append(string)
                    time_strings.append(time_line)

        if len(strings) <= 50:
            return reversed(strings)
        with open(self.filename, "wb") as f:
            for i in range(50):
                f.write(("\n%s" % time_strings[i - 50]).encode("utf-8"))
                for line in strings[i - 50].split("\n"):
                    f.write(("+%s\n" % line).encode("utf-8"))
        return reversed(strings)
    
    def store_string(self, string):
        if string.lower()[:4] == "exit":
            return 
        super(CliFileHistory, self).store_string(string)


class CliPrompt(object):
    def __init__(self, manager, session_config, executor):
        self.manager = manager
        self.session_config = session_config
        self.executor = executor

    def run(self):
        home_config_path = self.session_config.get_home()
        if not os.path.exists(home_config_path):
            os.mkdir(home_config_path)
        database_completer_words = [database["name"] for database in (self.session_config.get().get("databases") or [])
                                    if database["name"] not in ("-", "--")]
        sql_completer = WordCompleter(list(set(SQL_COMPLETER_WORDS + list(mysql_funcs.funcs.keys()) + database_completer_words)), ignore_case=True)
        history = CliFileHistory(os.path.join(home_config_path, "history"))
        session = PromptSession(
            lexer=PygmentsLexer(SqlLexer), completer=sql_completer, style=style, history=history
        )
        print("Python %s" % sys.version)
        print("Syncany-SQL %s -- Simple and easy-to-use sql execution engine" % version)
        lineno = 1
        while True:
            try:
                text = session.prompt("> ", multiline=Condition(lambda: self.check_complete(session.app.current_buffer.text)),
                                      cursor=CursorShape.BEAM)
                text = text.strip()
                if not text:
                    continue
                if text.lower()[:4] == "exit":
                    return
                if text[0] == "!":
                    self.run_system_cmd(text[1:])
                    continue
                if text[0] == "%":
                    self.run_magic_cmd(text[1:])
                    continue
                self.executor.run("cli", [SqlSegment(text, lineno)])
                try:
                    self.executor.execute()
                except Exception as e:
                    print(e.__class__.__name__ + ": " + str(e))
                finally:
                    lineno += 1
                    self.executor.runners.clear()
            except KeyboardInterrupt:
                continue  # Control-C pressed. Try again.
            except EOFError:
                return 130
            except Exception as e:
                print(e.__class__.__name__ + ": " + str(e))

    def check_complete(self, content):
        content = content.strip()
        if not content or content[:4] == "exit" or content[0] in ("!", "%"):
            return False
        sql_parser = SqlParser(content)
        try:
            sql_parser.read_util_complete()
        except EOFError:
            return True
        return False

    def run_system_cmd(self, cmd):
        os.system(cmd)

    def run_magic_cmd(self, cmd):
        if cmd[:3].lower() == "cd ":
            os.chdir(cmd[3:])
            self.executor.global_env_variables["cwd"] = os.getcwd()



# -*- coding: utf-8 -*-
# 2023/3/2
# create by: snower

import math
import random
from syncany.calculaters import typing_filter
from ...utils import ensure_number, ensure_int

@typing_filter(float)
def mysql_add(x, y):
    if x.__class__ is int or x.__class__ is float:
        if y.__class__ is int or y.__class__ is float:
            return x + y
    if x is None or y is None:
        return None
    return ensure_number(x) + ensure_number(y)

@typing_filter(float)
def mysql_sub(x, y):
    if x.__class__ is int or x.__class__ is float:
        if y.__class__ is int or y.__class__ is float:
            return x - y
    if x is None or y is None:
        return None
    return ensure_number(x) - ensure_number(y)

@typing_filter(float)
def mysql_mul(x, y):
    if x.__class__ is int or x.__class__ is float:
        if y.__class__ is int or y.__class__ is float:
            return x * y
    if x is None or y is None:
        return None
    return ensure_number(x) * ensure_number(y)

@typing_filter(float)
def mysql_div(x, y):
    if x.__class__ is int or x.__class__ is float:
        if y.__class__ is int or y.__class__ is float:
            return x / y
    if x is None or y is None:
        return None
    return ensure_number(x) / ensure_number(y)

@typing_filter(float)
def mysql_mod(x, y):
    if x.__class__ is int or x.__class__ is float:
        if y.__class__ is int or y.__class__ is float:
            return x % y
    if x is None or y is None:
        return None
    return ensure_number(x) % ensure_number(y)

@typing_filter(int)
def mysql_bitwiseand(x, y):
    if x is None or y is None:
        return None
    try:
        return x & y
    except:
        return ensure_int(x) & ensure_int(y)

@typing_filter(int)
def mysql_bitwiseor(x, y):
    if x is None or y is None:
        return None
    try:
        return x | y
    except:
        return ensure_int(x) | ensure_int(y)

@typing_filter(int)
def mysql_bitwisenot(x):
    if x is None:
        return None
    try:
        return ~ x
    except:
        return ~ ensure_int(x)

@typing_filter(int)
def mysql_bitwisexor(x, y):
    if x is None or y is None:
        return None
    try:
        return x ^ y
    except:
        return ensure_int(x) ^ ensure_int(y)

@typing_filter(int)
def mysql_bitwiserightshift(x, y):
    if x is None or y is None:
        return None
    try:
        return x >> y
    except:
        return ensure_int(x) >> ensure_int(y)

@typing_filter(int)
def mysql_bitwiseleftshift(x, y):
    if x is None or y is None:
        return None
    try:
        return x << y
    except:
        return ensure_int(x) << ensure_int(y)

@typing_filter(int)
def mysql_bit_and(x, y):
    if x is None or y is None:
        return None
    try:
        return x & y
    except:
        return ensure_int(x) & ensure_int(y)

@typing_filter(int)
def mysql_bit_or(x, y):
    if x is None or y is None:
        return None
    try:
        return x | y
    except:
        return ensure_int(x) | ensure_int(y)

@typing_filter(int)
def mysql_bit_xor(x, y):
    if x is None or y is None:
        return None
    try:
        return x ^ y
    except:
        return ensure_int(x) ^ ensure_int(y)

@typing_filter(float)
def mysql_abs(x):
    if x is None:
        return None
    try:
        return abs(x)
    except:
        return abs(ensure_number(x))

@typing_filter(float)
def mysql_sqrt(x):
    if x is None:
        return None
    try:
        return math.sqrt(x)
    except:
        return math.sqrt(ensure_number(x))

@typing_filter(float)
def mysql_exp(x):
    if x is None:
        return None
    try:
        return math.exp(x)
    except:
        return math.exp(ensure_number(x))

@typing_filter(float)
def mysql_pi():
    return math.pi

@typing_filter(float)
def mysql_ln(x):
    if x is None:
        return None
    try:
        return math.log(x, math.e)
    except:
        return math.log(ensure_number(x), math.e)

@typing_filter(float)
def mysql_log(x, base=10):
    if x is None:
        return None
    try:
        return math.log(x, base)
    except:
        return math.log(ensure_number(x), base)

@typing_filter(int)
def mysql_ceil(x):
    if x is None:
        return None
    try:
        return math.ceil(x)
    except:
        return math.ceil(ensure_number(x))

@typing_filter(int)
def mysql_ceiling(x):
    if x is None:
        return None
    try:
        return math.ceil(x)
    except:
        return math.ceil(ensure_number(x))

@typing_filter(int)
def mysql_floor(x):
    if x is None:
        return None
    try:
        return math.floor(x)
    except:
        return math.floor(ensure_number(x))

@typing_filter(float)
def mysql_rand():
    return random.random()

@typing_filter(float)
def mysql_round(x, y=2):
    if x is None or y is None:
        return None
    try:
        return round(x, y)
    except:
        return round(ensure_number(x), ensure_int(y))

@typing_filter(float)
def mysql_sign(x):
    if x is None:
        return None
    try:
        return 0 if x == 0 else (-1 if x < 0 else 1)
    except:
        x = ensure_number(x)
        return 0 if x == 0 else (-1 if x < 0 else 1)

@typing_filter(float)
def mysql_pow(x, y):
    if x is None or y is None:
        return None
    try:
        return math.pow(x, y)
    except:
        return math.pow(ensure_number(x), ensure_number(y))

@typing_filter(float)
def mysql_power(x, y):
    if x is None or y is None:
        return None
    try:
        return math.pow(x, y)
    except:
        return math.pow(ensure_number(x), ensure_number(y))

@typing_filter(float)
def mysql_sin(x):
    if x is None:
        return None
    try:
        return math.sin(x)
    except:
        return math.sin(ensure_number(x))

@typing_filter(float)
def mysql_asin(x):
    if x is None:
        return None
    try:
        return math.asin(x)
    except:
        return math.asin(ensure_number(x))

@typing_filter(float)
def mysql_cos(x):
    if x is None:
        return None
    try:
        return math.cos(x)
    except:
        return math.cos(ensure_number(x))

@typing_filter(float)
def mysql_acos(x):
    if x is None:
        return None
    try:
        return math.acos(x)
    except:
        return math.acos(ensure_number(x))

@typing_filter(float)
def mysql_tan(x):
    if x is None:
        return None
    try:
        return math.tan(x)
    except:
        return math.tan(ensure_number(x))

@typing_filter(float)
def mysql_atan(x):
    if x is None:
        return None
    try:
        return math.atan(x)
    except:
        return math.atan(ensure_number(x))

@typing_filter(float)
def mysql_greatest(*args):
    try:
        return max(*args)
    except:
        return max(*tuple([ensure_number(x) for x in args]))

@typing_filter(float)
def mysql_least(*args):
    try:
        return min(*args)
    except:
        return min(*tuple([ensure_number(x) for x in args]))


funcs = {key[6:]: value for key, value in globals().items() if key.startswith("mysql_")}


# -*- coding: utf-8 -*-
# 2023/2/13
# create by: snower

class SetCommandTasker(object):
    def __init__(self, config):
        self.config = config

    def start(self, name, executor, session_config, manager, arguments):
        key, is_global = self.config["key"], False
        if key[:6].lower() == "global":
            key, is_global = key[6:].strip(), True
        name = key.split(".")[0][1:]
        if name in ("databases", "imports", "sources", "defines", "variables", "options", "caches"):
            self.set_config(session_config, key[1:], is_global)
        elif name == "virtual_views":
            self.set_config(session_config, "databases." + key[1:], is_global)
        elif key[:7] == "@config":
            self.set_config(session_config, key[8:].strip(), is_global)
        else:
            try:
                self.set_env_variable(executor, key, self.parse_value(self.config["value"].strip()), is_global)
            except ValueError as e:
                if key[:1] == "@":
                    try:
                        executor.compile(name, "select " + self.config["value"].strip() + " into " + key)
                        return []
                    except Exception as e:
                        raise ValueError("unknown value: %s" % self.config["value"].strip())
                raise e
        return []

    def set_config(self, session_config, key, is_global=False):
        if is_global:
            session_config.global_config.set(key, self.parse_value(self.config["value"].strip()))
            session_config.merge()
            return
        session_config.set(key, self.parse_value(self.config["value"].strip()))

    def set_env_variable(self, executor, key, value, is_global=False):
        if is_global:
            executor.global_env_variables[key] = value
            return
        executor.env_variables[key] = value

    def parse_value(self, value):
        if not isinstance(value, str):
            return value
        if value[:1] in ('"', "'"):
            if value[:3] in ("'''", '"""'):
                return value[3:-3].strip()
            return value[1:-1].strip()
        if value.lower() == 'true':
            return True
        if value.lower() == 'false':
            return False
        if value.lower() == 'null':
            return None
        if value.isdigit() or (value[0] == "-" and value[1:].isdigit()):
            return int(value)
        value_info = value.split(".")
        if len(value_info) == 2 and (value_info[0].isdigit() or (value[0][0] == "-" and value[0][1:].isdigit())) \
                and value_info[1].isdigit():
            return float(value)
        raise ValueError("unknown value: %s" % value)

    def run(self, executor, session_config, manager):
        pass

    def terminate(self):
        pass


# -*- coding: utf-8 -*-
# 2023/7/3
# create by: snower

import datetime
import pytz
from syncany.utils import ensure_timezone, get_timezone
from .example import ExampleTestCase

gmt8 = pytz.timezone("Etc/GMT-8")


class TimeWindowExampleTestCase(ExampleTestCase):
    example_name = "time_window"

    def test_time_window(self):
        self.execute("time_window.sql")

        self.assert_value(2, "a", lambda value: isinstance(value, datetime.datetime), "data error")
        self.assert_value(2, "b", lambda value: isinstance(value, datetime.datetime), "data error")
        self.assert_value(2, "c", lambda value: isinstance(value, datetime.datetime), "data error")
        self.assert_value(2, "d", lambda value: isinstance(value, datetime.datetime), "data error")
        self.assert_value(2, "e", lambda value: isinstance(value, datetime.datetime), "data error")
        self.assert_value(2, "f", lambda value: isinstance(value, datetime.datetime), "data error")

        self.assert_value(4, "a", ensure_timezone(datetime.datetime(2023, 7, 3, 12, 24, 27, tzinfo=get_timezone())),
                          "data error")
        self.assert_value(4, "b", ensure_timezone(datetime.datetime(2023, 7, 3, 12, 24, 15, tzinfo=get_timezone())),
                          "data error")
        self.assert_value(4, "c", ensure_timezone(datetime.datetime(2023, 7, 3, 12, 24, tzinfo=get_timezone())),
                          "data error")
        self.assert_value(4, "d", ensure_timezone(datetime.datetime(2023, 7, 3, 12, 15, tzinfo=get_timezone())),
                          "data error")
        self.assert_value(4, "e", ensure_timezone(datetime.datetime(2023, 7, 3, 12, 0, tzinfo=get_timezone())),
                          "data error")
        self.assert_value(4, "f", ensure_timezone(datetime.datetime(2023, 7, 3, 0, 0, tzinfo=get_timezone())),
                          "data error")

        self.assert_value(7, "a", ensure_timezone(datetime.datetime(2023, 7, 3, 12, 24, 30, tzinfo=get_timezone())),
                          "data error")
        self.assert_value(7, "b", ensure_timezone(datetime.datetime(2023, 7, 3, 12, 25, tzinfo=get_timezone())),
                          "data error")
        self.assert_value(7, "c", ensure_timezone(datetime.datetime(2023, 7, 3, 12, 27, tzinfo=get_timezone())),
                          "data error")
        self.assert_value(7, "d", ensure_timezone(datetime.datetime(2023, 7, 3, 13, 0, tzinfo=get_timezone())),
                          "data error")
        self.assert_value(7, "e", ensure_timezone(datetime.datetime(2023, 7, 3, 18, 0, tzinfo=get_timezone())),
                          "data error")
        self.assert_value(7, "f", ensure_timezone(datetime.datetime(2023, 7, 6, 0, 0, tzinfo=get_timezone())),
                          "data error")

        self.assert_value(10, "TIME_WINDOW('15M', create_time)",
                          ensure_timezone(datetime.datetime(2018, 9, 15, 12, 15, tzinfo=gmt8)), "data error")
        self.assert_value(10, "COUNT(*)", 3, "data error")
        self.assert_value(10, "TIME_WINDOW('15M', create_time)",
                          ensure_timezone(datetime.datetime(2018, 9, 15, 14, 15, tzinfo=gmt8)), "data error", index=1)
        self.assert_value(10, "COUNT(*)", 3, "data error", index=1)
        self.assert_value(10, "TIME_WINDOW('15M', create_time)",
                          ensure_timezone(datetime.datetime(2018, 9, 13, 16, 45, tzinfo=gmt8)), "data error", index=2)
        self.assert_value(10, "COUNT(*)", 3, "data error", index=2)
        self.assert_value(10, "TIME_WINDOW('15M', create_time)",
                          ensure_timezone(datetime.datetime(2018, 9, 13, 17, 0, tzinfo=gmt8)), "data error", index=3)
        self.assert_value(10, "COUNT(*)", 3, "data error", index=3)
        self.assert_value(10, "TIME_WINDOW('15M', create_time)",
                          ensure_timezone(datetime.datetime(2018, 9, 14, 16, 15, tzinfo=gmt8)), "data error", index=4)
        self.assert_value(10, "COUNT(*)", 3, "data error", index=4)
        self.assert_value(10, "TIME_WINDOW('15M', create_time)",
                          ensure_timezone(datetime.datetime(2020, 2, 25, 15, 45, tzinfo=gmt8)), "data error", index=5)
        self.assert_value(10, "COUNT(*)", 1, "data error", index=5)
        self.assert_value(10, "TIME_WINDOW('15M', create_time)",
                          ensure_timezone(datetime.datetime(2020, 2, 25, 16, 0, tzinfo=gmt8)), "data error", index=6)
        self.assert_value(10, "COUNT(*)", 2, "data error", index=6)
        self.assert_value(10, "TIME_WINDOW('15M', create_time)",
                          ensure_timezone(datetime.datetime(2020, 2, 25, 16, 45, tzinfo=gmt8)), "data error", index=7)
        self.assert_value(10, "COUNT(*)", 1, "data error", index=7)
        self.assert_value(10, "TIME_WINDOW('15M', create_time)",
                          ensure_timezone(datetime.datetime(2020, 5, 16, 15, 15, tzinfo=gmt8)), "data error", index=8)
        self.assert_value(10, "COUNT(*)", 1, "data error", index=8)
        self.assert_value(10, "TIME_WINDOW('15M', create_time)",
                          ensure_timezone(datetime.datetime(2020, 5, 17, 15, 0, tzinfo=gmt8)), "data error", index=9)
        self.assert_value(10, "COUNT(*)", 2, "data error", index=9)
